# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import gc
import gzip
import os
from glob import glob
import warnings
try:
    import cv2
    IS_CV2 = True
except ImportError:
    IS_CV2 = False
try:
    import h5py
    IS_H5PY = True
except ImportError:
    IS_H5PY = False


import numpy as np

from ..core import (Points, Profile, ScalarField, SpatialScalarFields,
                    SpatialVectorFields, TemporalScalarFields,
                    TemporalVectorFields,
                    VectorField)
from ..vortex_detection import CritPoints
from ..utils import ProgressCounter, make_unit
from ..utils.types import ARRAYTYPES, STRINGTYPES


try:
    import pickle as pickle
except:
    import pickle
import scipy.io as spio
import imageio
from os import path
import re


def check_path(filepath, newfile=False):
    """
    Normalize and check the validity of the given path to feed importation
    functions.
    """
    # check
    if not isinstance(filepath, STRINGTYPES):
        raise TypeError()
    if not isinstance(newfile, bool):
        raise ValueError()
    # normalize
    filepath = path.normpath(filepath)
    # check validity
    if newfile:
        filepath, filename = path.split(filepath)
    if not path.exists(filepath):
        # split the path (to check existing part)
        path_compos = []
        p = filepath
        while True:
            p, f = path.split(p)
            if f != "":
                path_compos.append(f)
            else:
                if p != "":
                    path_compos.append(p)
                break
        # check validity recursively
        valid_path = ""
        while True:
            if len(path_compos) == 0:
                break
            new_dir = path_compos.pop()
            if new_dir == "":
                break
            new_tested_path = path.join(valid_path, new_dir)
            if not path.exists(new_tested_path):
                if valid_path == "":
                    valid_path = os.getcwd()
                err_mess = r"'{}' directory/file not found in '{}' directory." \
                           .format(new_dir, valid_path)
                raise FileNotFoundError(err_mess)
            valid_path = new_tested_path
    # returning
    if newfile:
        filepath = path.join(filepath, filename)
    return filepath


def find_file_in_path(regs, dirpath, ask=False):
    """
    Search recursively for a folder containing files matching a regular
    expression, in the given root folder.

    Parameters
    ----------
    exts : list of string
        List of regular expressions
    dirpath : string
        Root path to search from
    ask : bool
        If 'True', ask for the wanted folder and return only this one.

    Returns
    -------
    folders : list of string
        List of folder containings wanted files.
    """
    # check
    if not path.isdir(dirpath):
        raise ValueError()
    if not isinstance(regs, ARRAYTYPES):
        raise TypeError()
    regs = np.array(regs, dtype=str)
    #
    dir_paths = []
    # recursive loop on folders
    for root, dirs, files in os.walk(dirpath):
        for f in files:
            match = np.any([re.match(reg, f) for reg in regs])
            if match:
                break
        if match:
            dir_paths.append(root)
    # choose
    if ask and len(dir_paths) > 1:
        print(("{} folders found :".format(len(dir_paths))))
        for i, p in enumerate(dir_paths):
            print(("  {} :  {}".format(i + 1, p)))
        rep = 0
        while rep not in np.arange(1, len(dir_paths)+1):
            rep = input("Want to go with wich one ?\n")
            try:
                rep = int(rep)
            except:
                pass
        dir_paths = [dir_paths[rep - 1]]
    # return
    return dir_paths


def matlab_parser(obj, name):
    classic_types = (int, float, str, bool)
    array_types = (list, float)
    if obj is None:
        return {name: "None"}
    elif isinstance(obj, classic_types):
        return {name: obj}
    elif isinstance(obj, array_types):
        simple = True
        for val in obj:
            if not isinstance(val, (classic_types, array_types)):
                simple = False
        if simple:
            return {name: obj}
        else:
            return {name: [matlab_parser(obj[i], f"obj{i}")
                           for i in range(len(obj))]}
            raise IOError("Matlab can't handle this kind of variable")
    elif isinstance(obj, Points):
        xy = obj.xy
        xy = [list(comp) for comp in xy]
        dic = matlab_parser(list(obj.v), 'v')
        dic.update(matlab_parser(xy, 'xy'))
        dic.update(matlab_parser(obj.unit_x.strUnit()[1:-1], 'unit_x'))
        dic.update(matlab_parser(obj.unit_y.strUnit()[1:-1], 'unit_y'))
        dic.update(matlab_parser(obj.unit_v.strUnit()[1:-1], 'unit_v'))
        return {name: dic}
    elif isinstance(obj, Profile):
        mask = [int(val) for val in obj.mask]
        dic = matlab_parser(list(obj.x), 'x')
        dic.update(matlab_parser(list(obj.y), 'y'))
        dic.update(matlab_parser(mask, 'mask'))
        dic.update(matlab_parser(obj.unit_x.strUnit()[1:-1], 'unit_x'))
        dic.update(matlab_parser(obj.unit_y.strUnit()[1:-1], 'unit_y'))
        return {name: dic}
    elif isinstance(obj, VectorField):
        x = obj.axe_x
        y = obj.axe_y
        X, Y = np.meshgrid(x, y, indexing='ij')
        X = [list(comp) for comp in X]
        Y = [list(comp) for comp in Y]
        comp_x = obj.comp_x
        comp_x = [list(comp) for comp in comp_x]
        comp_y = obj.comp_y
        comp_y = [list(comp) for comp in comp_y]
        mask = obj.mask
        mask = [list(comp) for comp in mask]
        dic = matlab_parser(list(obj.axe_x), 'axe_x')
        dic.update(matlab_parser(list(obj.axe_y), 'axe_y'))
        dic.update(matlab_parser(list(X), 'X'))
        dic.update(matlab_parser(list(Y), 'Y'))
        dic.update(matlab_parser(comp_x, 'comp_x'))
        dic.update(matlab_parser(comp_y, 'comp_y'))
        dic.update(matlab_parser(mask, 'mask'))
        dic.update(matlab_parser(obj.unit_x.strUnit()[1:-1], 'unit_x'))
        dic.update(matlab_parser(obj.unit_y.strUnit()[1:-1], 'unit_y'))
        dic.update(matlab_parser(obj.unit_values.strUnit()[1:-1],
                                 'unit_values'))
        return {name: dic}
    elif isinstance(obj, ScalarField):
        x = obj.axe_x
        y = obj.axe_y
        X, Y = np.meshgrid(x, y, indexing='ij')
        X = [list(comp) for comp in X]
        Y = [list(comp) for comp in Y]
        values = obj.values
        values = [list(comp) for comp in values]
        mask = obj.mask
        mask = [list(comp) for comp in mask]
        dic = matlab_parser(list(obj.axe_x), 'axe_x')
        dic.update(matlab_parser(list(obj.axe_y), 'axe_y'))
        dic.update(matlab_parser(list(X), 'X'))
        dic.update(matlab_parser(list(Y), 'Y'))
        dic.update(matlab_parser(values, 'values'))
        dic.update(matlab_parser(mask, 'mask'))
        dic.update(matlab_parser(obj.unit_x.strUnit()[1:-1], 'unit_x'))
        dic.update(matlab_parser(obj.unit_y.strUnit()[1:-1], 'unit_y'))
        dic.update(matlab_parser(obj.unit_values.strUnit()[1:-1],
                                 'unit_values'))
        return {name: dic}
    elif isinstance(obj, CritPoints):
        eps = obj.current_epsilon
        if eps is not None:
            foc_traj = list(obj.foc_traj)
            foc_c_traj = list(obj.foc_c_traj)
            node_i_traj = list(obj.node_i_traj)
            node_o_traj = list(obj.node_o_traj)
            sadd_traj = list(obj.sadd_traj)
        else:
            foc_traj = None
            foc_c_traj = None
            node_i_traj = None
            node_o_traj = None
            sadd_traj = None
        foc = list(obj.foc)
        foc_c = list(obj.foc_c)
        node_i = list(obj.node_i)
        node_o = list(obj.node_o)
        sadd = list(obj.sadd)
        times = list(obj.times)
        unit_time = obj.unit_time.strUnit()
        unit_x = obj.unit_x.strUnit()
        unit_y = obj.unit_y.strUnit()
        dic = matlab_parser(eps, "epsilon")
        dic.update(matlab_parser(foc, "foc"))
        dic.update(matlab_parser(foc_traj, "foc_traj"))
        dic.update(matlab_parser(foc_c, "foc_c"))
        dic.update(matlab_parser(foc_c_traj, "foc_c_traj"))
        dic.update(matlab_parser(node_i, "node_i"))
        dic.update(matlab_parser(node_o, "node_o"))
        dic.update(matlab_parser(node_i_traj, "node_i_traj"))
        dic.update(matlab_parser(node_o_traj, "node_o_traj"))
        dic.update(matlab_parser(sadd, "sadd"))
        dic.update(matlab_parser(sadd_traj, "sadd_traj"))
        dic.update(matlab_parser(times, "times"))
        dic.update(matlab_parser(unit_time, "unit_time"))
        dic.update(matlab_parser(unit_x, "unit_x"))
        dic.update(matlab_parser(unit_y, "unit_y"))
        return {name: dic}
    else:
        raise IOError("Can't parser that : \n {}".format(obj))


def import_from_file(filepath, **kw):
    """
    Load and return an object from the specified file using the JSON
    format.
    Additionnals arguments for the JSON decoder may be set with the **kw
    argument. Such as'encoding' (to change the file
    encoding, default='utf-8').

    Parameters
    ----------
    filepath : string
        Path specifiing the file to load.
    full_import : boolean
        If 'True', everything is charged in memory, else, data are loaded in
        memory when they are needed.
    """
    # getting/guessing wanted files
    filepath = check_path(filepath)
    extension = path.splitext(filepath)[1]
    # importing file
    if extension == ".imt":
        gc.disable()
        with open(filepath, 'rb') as f:
            obj = pickle.load(f)
        gc.enable()
    elif extension == ".cimt":
        gc.disable()
        with gzip.open(filepath, 'rb') as f:
            obj = pickle.load(f)
        gc.enable()
    else:
        raise IOError("File is not readable "
                      "(unknown extension : {})".format(extension))
    # For backward support
    attr_name = f"_values_dtype"
    if isinstance(obj, (VectorField, ScalarField)):
        if not hasattr(obj, attr_name):
            setattr(obj, attr_name, float)
    return obj


def export_to_file(obj, filepath, compressed=True, **kw):
    """
    Write the object in the specified file.
    Additionnals arguments for the JSON encoder may be set with the **kw
    argument.
    If existing, specified file will be truncated. If not, it will
    be created.

    Parameters
    ----------
    obj :
        Object to store (common and IMT objects are supported).
    filepath : string
        Path specifiing where to save the object.
    compressed : boolean, optional
        If 'True' (default), the file is compressed using gzip.
    """
    # checking parameters coherence
    filepath = check_path(filepath, newfile=True)
    if not isinstance(compressed, bool):
        # creating/filling up the file
        raise TypeError("'compressed' must be a boolean")
    if compressed:
        gc.disable()
        if path.splitext(filepath)[1] != ".cimt":
            filepath = filepath + ".cimt"
        f = gzip.open(filepath, 'wb')
        pickle.dump(obj, f, protocol=-1)
        f.close()
        gc.enable()
    else:
        gc.disable()
        if path.splitext(filepath)[1] != ".imt":
            filepath = filepath + ".imt"
        f = open(filepath, 'wb')
        pickle.dump(obj, f, protocol=-1)
        f.close()
        gc.enable()


def imts_to_imt(imts_path, imt_path, kind):
    """
    Concatenate some .imt files to one .imt file.

    Parameters
    ----------
    imts_path : string
        Path to the .imt files
    imt_path : string
        Path to store the new imt file.
    kind : string
        Kind of object for the new imt file
        (can be 'TSF' for TemporalScalarFields, 'SSF' for SpatialScalarFields,
        'TVF' for TemporalVectorFields, 'SVF' for SpatialVectorFields)
    """
    # check parameters
    imts_path = check_path(imts_path)
    imt_path = check_path(imt_path, newfile=True)
    if not isinstance(kind, STRINGTYPES):
        raise TypeError()
    # getting paths
    paths = glob(imts_path + "/*")
    # getting data type
    if kind == 'TSF':
        imts_type = 'SF'
        fields = TemporalScalarFields()
    elif kind == 'SSF':
        imts_type = 'SF'
        fields = SpatialScalarFields()
    elif kind == 'TVF':
        imts_type = 'VF'
        fields = TemporalVectorFields()
    elif kind == 'SVF':
        imts_type = 'VF'
        fields = SpatialVectorFields()
    else:
        raise ValueError()
    # importing data
    for tmp_path in paths:
        basename = path.basename(tmp_path)
        name, ext = path.splitext(basename)
        if ext in ['.imt', '.cimt']:
            field = import_from_file(tmp_path)
            if imts_type == 'SF' and not isinstance(field, ScalarField):
                continue
            elif imts_type == 'VF' and not isinstance(field, VectorField):
                continue
            fields.add_field(field)
    # saving data
    export_to_file(fields, imt_path)


def import_from_matlab(filepath, obj, show_struct=False, **kwargs):
    """
    Import data from a matlab (.m) file.

    Data should be a dictionnary.

    Parameters
    ----------
    filepath : string
        Path of the matlab file
    obj : string in ['ScalarField', 'VectorField', 'Profile',
                     'Points']
        Kind of object to import to.
    show_struct: boolean
       If True, just show the structure of the file
       (and return it, so you can play with it)
    kwargs :
       Rest of the keyword arguments should indicate where to find
       the necessary information in the matlab dictionnary.

    Example
    -------
    With a matlab file containing a dictionnary with the
    following entries: 'x', 'x_unit', 'y', 'y_unit', 'u', 'v'
    >>> vf = import_from_matlab('data.m', 'VectorField',
    ...                         axe_x='x', axe_y='y',
    ...                         unit_x='x_unit', unit_y='y_unit',
    ...                         comp_x='u', comp_y='v')
    """
    if not IS_H5PY:
        raise Exception('h5py needs to be installed to import from matlab.')
    filepath = check_path(filepath)
    # Get matlab dictionnary
    try:
        data = spio.loadmat(filepath)
        is_h5py = False
    except NotImplementedError:
        data = h5py.File(filepath, 'r')
        is_h5py = True
    if isinstance(data, np.ndarray):
        keys = list(data.dtype.fields.keys())
    else:
        keys = list(data.keys())
    # pivmat !
    dataf = None
    if 'Data' in keys and 'P' in keys:
        dataf = data
        data = dataf['Data']
        if isinstance(data, np.ndarray):
            keys = list(data.dtype.fields.keys())
        else:
            keys = list(data.keys())
    # Create object
    if obj == 'ScalarField':
        res = ScalarField
        default_dic = {'axe_x': 'axe_x', 'axe_y': 'axe_y', 'unit_x': 'unit_x',
                       'unit_y': 'unit_y', 'values': 'values', 'mask': 'mask',
                       'unit_values': 'unit_values'}
    elif obj == 'VectorField':
        res = VectorField
        default_dic = {'axe_x': 'axe_x', 'axe_y': 'axe_y', 'comp_x': 'comp_x',
                       'comp_y': 'comp_y', 'mask': 'mask', 'unit_x': 'unit_x',
                       'unit_y': 'unit_y', 'unit_values': 'unit_values'}
    elif obj == 'Profile':
        res = Profile
        default_dic = {'x': 'x', 'y': 'y', 'mask': 'mask', 'unit_x': 'unit_x',
                       'unit_y': 'unit_y'}
    elif obj == 'Points':
        res = Points
        default_dic = {'xy': 'xy', 'v': 'v', 'unit_x': 'unit_x',
                       'unit_y': 'unit_y', 'unit_v': 'unit_v'}
    else:
        raise Exception()
    default_dic.update(kwargs)
    kwargs = default_dic
    # Just show the structure
    if show_struct:
        if is_h5py:
            print(f"Available keys in {filepath}:\n{keys})")
        else:
            print(f"Available keys in {filepath}:"
                  f"\n{keys}")
        return data
    # Create building dict
    build_dic = {}
    for key, entry in kwargs.items():
        if entry in keys:
            tdata = data[entry]
            while not isinstance(tdata, str) and len(tdata) == 1:
                tdata = tdata[0]
            print(tdata)
            if key in ['unit_x', 'unit_y', 'unit_values']:
                if isinstance(tdata, str):
                    build_dic[key] = tdata
                else:
                    build_dic[key] = ''.join(chr(n) for n in tdata)
            else:
                build_dic[key] = np.array(tdata)
    if is_h5py and dataf is not None:
        dataf.close()
    # Fix unities
    for key, item in build_dic.items():
        try:
            if len(item) == 1:
                build_dic[key] = build_dic[key][0]
        except TypeError:
            pass
    # Fill the object
    if obj in ['ScalarField', 'VectorField']:
        ress = res()
        ress.import_from_arrays(**build_dic)
    else:
        ress = res(**build_dic)
    # Return
    return ress


def export_to_matlab(obj, filepath, **kw):
    filepath = check_path(filepath, newfile=True)
    dic = matlab_parser(obj, "tmp")
    spio.savemat(filepath, dic["tmp"], **kw)


def export_to_vtk(obj, filepath, axis=None, **kw):
    """
    Export the field to a .vtk file, for Mayavi use.

    Parameters
    ----------
    filepath : string
        Path where to write the vtk file.
    axis : tuple of strings
        By default, field axe are set to (x,y), if you want
        different axis, you have to specified them here.
        For example, "('z', 'y')", put the x field axis values
        in vtk z axis, and y field axis in y vtk axis.
    line : boolean (only for Points object)
        If 'True', lines between points are writen instead of points.
    """
    if isinstance(obj, ScalarField):
        __export_sf_to_vtk(obj, filepath, axis)
    elif isinstance(obj, VectorField):
        __export_vf_to_vtk(obj, filepath, axis)
    elif isinstance(obj, Points):
        __export_pts_to_vtk(obj, filepath, **kw)
    else:
        raise TypeError("Cannot (yet) export this kind of object to vtk")


def __export_pts_to_vtk(pts, filepath, axis=None, line=False):
    """
    Export the Points object to a .vtk file, for Mayavi use.

    Parameters
    ----------
    pts : Point object
        .
    filepath : string
        Path where to write the vtk file.
    axis : tuple of strings, optional
        By default, points field axe are set to (x,y), if you want
        different axis, you have to specified them here.
        For example, "('z', 'y')", put the x points field axis values
        in vtk z axis, and y points field axis in y vtk axis.
    line : boolean, optional
        If 'True', lines between points are writen instead of points.
    """
    try:
        import pyvtk
    except ImportError:
        raise Exception("You need to install pyvtk to use this "
                        "functionnality")
    if not path.exists(path.dirname(filepath)):
        raise ValueError("'filepath' is not a valid path")
    if axis is None:
        axis = ('x', 'y')
    if not isinstance(axis, ARRAYTYPES):
        raise TypeError("'axis' must be a 2x1 tuple")
    if not isinstance(axis[0], STRINGTYPES) \
            or not isinstance(axis[1], STRINGTYPES):
        raise TypeError("'axis' must be a 2x1 tuple of strings")
    if not axis[0] in ['x', 'y', 'z'] or not axis[1] in ['x', 'y', 'z']:
        raise ValueError("'axis' strings must be 'x', 'y' or 'z'")
    if axis[0] == axis[1]:
        raise ValueError("'axis' strings must be different")
    if not isinstance(line, bool):
        raise TypeError("'line' must be a boolean")
    v = pts.v
    x = pts.xy[:, 0]
    y = pts.xy[:, 1]
    if v is None:
        v = np.zeros(pts.xy.shape[0])
    point_data = pyvtk.PointData(pyvtk.Scalars(v, 'Points values'))
    x_vtk = np.zeros(pts.xy.shape[0])
    y_vtk = np.zeros(pts.xy.shape[0])
    z_vtk = np.zeros(pts.xy.shape[0])
    if axis[0] == 'x':
        x_vtk = x
    elif axis[0] == 'y':
        y_vtk = x
    else:
        z_vtk = x
    if axis[1] == 'x':
        x_vtk = y
    elif axis[1] == 'y':
        y_vtk = y
    else:
        z_vtk = y
    pts = list(zip(x_vtk, y_vtk, z_vtk))
    vertex = np.arange(x_vtk.shape[0])
    if line:
        grid = pyvtk.UnstructuredGrid(pts, line=vertex)
    else:
        grid = pyvtk.UnstructuredGrid(pts, vertex=vertex)
    data = pyvtk.VtkData(grid, 'Scalar Field from python', point_data)
    data.tofile(filepath)


def __export_sf_to_vtk(obj, filepath, axis=None):
    """
    Export the scalar field to a .vtk file, for Mayavi use.

    Parameters
    ----------
    filepath : string
        Path where to write the vtk file.
    axis : tuple of strings
        By default, scalar field axe are set to (x,y), if you want
        different axis, you have to specified them here.
        For example, "('z', 'y')", put the x scalar field axis values
        in vtk z axis, and y scalar field axis in y vtk axis.
    """
    import pyvtk
    if not path.exists(path.dirname(filepath)):
        raise ValueError("'filepath' is not a valid path")
    if axis is None:
        axis = ('x', 'y')
    if not isinstance(axis, ARRAYTYPES):
        raise TypeError("'axis' must be a 2x1 tuple")
    if not isinstance(axis[0], STRINGTYPES) \
            or not isinstance(axis[1], STRINGTYPES):
        raise TypeError("'axis' must be a 2x1 tuple of strings")
    if not axis[0] in ['x', 'y', 'z'] or not axis[1] in ['x', 'y', 'z']:
        raise ValueError("'axis' strings must be 'x', 'y' or 'z'")
    if axis[0] == axis[1]:
        raise ValueError("'axis' strings must be different")
    V = obj.values.flatten()
    x = obj.axe_x
    y = obj.axe_y
    point_data = pyvtk.PointData(pyvtk.Scalars(V, 'Scalar Field'))
    x_vtk = 0.
    y_vtk = 0.
    z_vtk = 0.
    if axis[0] == 'x':
        x_vtk = x
    elif axis[0] == 'y':
        y_vtk = x
    else:
        z_vtk = x
    if axis[1] == 'x':
        x_vtk = y
    elif axis[1] == 'y':
        y_vtk = y
    else:
        z_vtk = y
    grid = pyvtk.RectilinearGrid(x_vtk, y_vtk, z_vtk)
    data = pyvtk.VtkData(grid, 'Scalar Field from python', point_data)
    data.tofile(filepath)


def __export_vf_to_vtk(obj, filepath, axis=None):
    """
    Export the vector field to a .vtk file, for Mayavi use.

    Parameters
    ----------
    filepath : string
        Path where to write the vtk file.
    axis : tuple of strings
        By default, scalar field axe are set to (x,y), if you want
        different axis, you have to specified them here.
        For example, "('z', 'y')", put the x scalar field axis values
        in vtk z axis, and y scalar field axis in y vtk axis.
    """
    import pyvtk
    if not path.exists(path.dirname(filepath)):
        raise ValueError("'filepath' is not a valid path")
    if axis is None:
        axis = ('x', 'y')
    if not isinstance(axis, ARRAYTYPES):
        raise TypeError("'axis' must be a 2x1 tuple")
    if not isinstance(axis[0], STRINGTYPES) \
            or not isinstance(axis[1], STRINGTYPES):
        raise TypeError("'axis' must be a 2x1 tuple of strings")
    if not axis[0] in ['x', 'y', 'z'] or not axis[1] in ['x', 'y', 'z']:
        raise ValueError("'axis' strings must be 'x', 'y' or 'z'")
    if axis[0] == axis[1]:
        raise ValueError("'axis' strings must be different")
    Vx, Vy = obj.comp_x, obj.comp_y
    Vx = Vx.flatten()
    Vy = Vy.flatten()
    x, y = obj.axe_x, obj.axe_y
    x_vtk = 0.
    y_vtk = 0.
    z_vtk = 0.
    vx_vtk = np.zeros(Vx.shape)
    vy_vtk = np.zeros(Vx.shape)
    vz_vtk = np.zeros(Vx.shape)
    if axis[0] == 'x':
        x_vtk = x
        vx_vtk = Vx
    elif axis[0] == 'y':
        y_vtk = x
        vy_vtk = Vx
    else:
        z_vtk = x
        vz_vtk = Vx
    if axis[1] == 'x':
        x_vtk = y
        vx_vtk = Vy
    elif axis[1] == 'y':
        y_vtk = y
        vy_vtk = Vy
    else:
        z_vtk = y
        vz_vtk = Vy
    vect = list(zip(vx_vtk, vy_vtk, vz_vtk))
    point_data = pyvtk.PointData(pyvtk.Vectors(vect, "Vector field"))
    grid = pyvtk.RectilinearGrid(x_vtk, y_vtk, z_vtk)
    data = pyvtk.VtkData(grid, 'Vector Field from python', point_data)
    data.tofile(filepath)


def _get_imx_buffers(filename):
    """
    Return the buffers stored in the given file.
    """
    try:
        import ReadIM
    except ModuleNotFoundError:
        raise Exception("You need the 'ReadIM' module to import from davis"
                        " files")
    vbuff, vatts = ReadIM.extra.get_Buffer_andAttributeList(filename)
    arrays, vbuff2 = ReadIM.extra.buffer_as_array(vbuff)
    arrays = np.array(arrays.transpose((0, 2, 1)))
    atts = ReadIM.extra.att2dict(vatts)
    fmt = vbuff.image_sub_type
    vectorGrid = vbuff.vectorGrid
    ReadIM.DestroyBuffer(vbuff)
    ReadIM.DestroyBuffer(vbuff2)
    return fmt, vectorGrid, arrays, atts


def import_from_IM7(filename, infos=False):
    """
    Import a scalar field from a .IM7 file.

    Parameters
    ----------
    filename : string
        Path to the IM7 file.
    infos : boolean, optional
        If 'True', also return a dictionary with informations on the im7
    """
    if not isinstance(filename, STRINGTYPES):
        raise TypeError("'filename' must be a string")
    if not path.exists(filename):
        raise ValueError("I did not find your file, boy")
    _, ext = path.splitext(filename)
    if not (ext == ".im7" or ext == ".IM7"):
        raise ValueError("I need the file to be an IM7 file (not a {} file)"
                         .format(ext))
    # Importing from buffer
    fmt, vectorGrid, v_array, atts = _get_imx_buffers(filename)
    if v_array.shape[0] == 2:
        mask = v_array[0][:, ::1]
        values = v_array[1][:, ::1]
    elif v_array.shape[0] == 1:
        values = v_array[0][:, ::1]
        mask = np.zeros(values.shape, dtype=bool)
    # Values and Mask
    scale_i = atts['_SCALE_I']
    scale_i = scale_i.split("\n")
    scale_val = scale_i[0].split(' ')
    unit_values = scale_i[1]
    try:
        values *= int(scale_val[0])
        values += int(scale_val[1])
    except ValueError:
        values *= float(scale_val[0])
        values += float(scale_val[1])
    # X
    scale_x = atts['_SCALE_X']
    scale_x = scale_x.split("\n")
    unit_x = scale_x[1]
    scale_val = scale_x[0].split(' ')
    x_init = float(scale_val[1])
    dx = float(scale_val[0])
    len_axe_x = values.shape[0]
    if dx < 0:
        axe_x = x_init + np.arange(len_axe_x - 1, -1, -1)*dx
        values = values[::-1, :]
        mask = mask[::-1, :]
    else:
        axe_x = x_init + np.arange(len_axe_x)*dx
    # Y
    scale_y = atts['_SCALE_Y']
    scale_y = scale_y.split("\n")
    unit_y = scale_y[1]
    scale_val = scale_y[0].split(' ')
    y_init = float(scale_val[1])
    dy = float(scale_val[0])
    len_axe_y = values.shape[1]
    if dy < 0:
        axe_y = y_init + np.arange(len_axe_y - 1, -1, -1)*dy
        values = values[:, ::-1]
        mask = mask[:, ::-1]
    else:
        axe_y = y_init + np.arange(len_axe_y)*dy
    # returning
    tmpsf = ScalarField()
    tmpsf.import_from_arrays(axe_x=axe_x, axe_y=axe_y, values=values,
                             mask=mask,
                             unit_x=unit_x, unit_y=unit_y,
                             unit_values=unit_values)
    if infos:
        return tmpsf, atts
    else:
        return tmpsf


def import_from_IM7s(fieldspath, kind='TSF', fieldnumbers=None, incr=1):
    """
    Import scalar fields from .IM7 files.
    'fieldspath' should be a tuple of path to im7 files.
    All im7 file present in the folder are imported.

    Parameters
    ----------
    fieldspath : string or tuple of string
    kind : string, optional
        Kind of object to create with IM7 files.
        (can be 'TSF' for TemporalScalarFields
        or 'SSF' for SpatialScalarFields).
    fieldnumbers : 2x1 tuple of int
        Interval of fields to import, default is all.
    incr : integer
        Incrementation between fields to take. Default is 1, meaning all
        fields are taken.

    """
    # check parameters
    if isinstance(fieldspath, ARRAYTYPES):
        if not isinstance(fieldspath[0], STRINGTYPES):
            raise TypeError("'fieldspath' must be a string or a tuple of"
                            " string")
        fieldspaths = np.asarray(fieldspath, dtype=str)
    elif isinstance(fieldspath, STRINGTYPES):
        fieldspath = check_path(fieldspath)
        paths = np.asarray([f for f in glob(path.join(fieldspath, '*'))
                            if path.splitext(f)[-1] in ['.im7', '.IM7']],
                           dtype=str)
        # if no file found, search recursively
        if len(paths) == 0:
            poss_paths = find_file_in_path(['.*.im7', '.*.IM7'], fieldspath,
                                           ask=True)
            if len(poss_paths) == 0:
                raise ValueError()
            paths = np.asarray([f
                                for f in glob(path.join(poss_paths[0], '*'))
                                if path.splitext(f)[-1] in ['.im7', '.IM7']],
                               dtype=str)
        # Sort path by numbers
        filenames = [path.basename(p) for p in paths]
        ind_sort = np.argsort(filenames)
        fieldspaths = paths[ind_sort]
    else:
        raise TypeError()
    if fieldnumbers is not None:
        if not isinstance(fieldnumbers, ARRAYTYPES):
            raise TypeError("'fieldnumbers' must be a 2x1 array")
        if not len(fieldnumbers) == 2:
            raise TypeError("'fieldnumbers' must be a 2x1 array")
        if not isinstance(fieldnumbers[0], int) \
                or not isinstance(fieldnumbers[1], int):
            raise TypeError("'fieldnumbers' must be an array of integers")
    else:
        fieldnumbers = [0, len(fieldspaths)]
    if not isinstance(incr, int):
        raise TypeError("'incr' must be an integer")
    if incr <= 0:
        raise ValueError("'incr' must be positive")
    # Import
    if kind == 'TSF':
        fields = TemporalScalarFields()
    elif kind == 'SSF':
        fields = SpatialScalarFields()
    else:
        raise ValueError()
    start = fieldnumbers[0]
    end = fieldnumbers[1]
    t = 0.
    # loop on files
    for p in fieldspaths[start:end:incr]:
        tmp_sf, infos = import_from_IM7(p, infos=True)
        try:
            dt = infos['FrameDt0'].split()
        except KeyError:
            dt = [1., ""]
        unit_time = make_unit(dt[1])
        dt = float(dt[0])
        t += dt*incr
        if kind == 'TSF':
            fields.add_field(tmp_sf, t, unit_time)
        else:
            fields.add_field(tmp_sf)
    return fields


def import_from_VC7(filename, infos=False, add_fields=False):
    """
    Import a vector field or a velocity field from a .VC7 file

    Parameters
    ----------
    filename : string
        Path to the file to import.
    infos : boolean, optional
        If 'True', also return a dictionary with informations on the im7
    add_fields : boolean, optional
        If 'True', also return a tuple containing additional fields
        contained in the vc7 field (peak ratio, correlation value, ...)
    """
    # check parameters
    filename = check_path(filename)
    _, ext = path.splitext(filename)
    if not (ext == ".vc7" or ext == ".VC7"):
        raise ValueError("'filename' must be a vc7 file")
    # Importing from buffer
    fmt, vectorGrid, v_array, atts = _get_imx_buffers(filename)
    # Values and Mask
    if fmt == 2:
        Vx = v_array[0]
        Vy = v_array[1]
        mask = np.zeros(Vx.shape, dtype=bool)
    elif fmt == 3 or fmt == 1:
        mask = np.logical_not(v_array[0])
        mask2 = np.logical_not(v_array[9])
        mask = np.logical_or(mask, mask2)
        Vx = v_array[1]
        Vy = v_array[2]
    mask = np.logical_or(mask, np.logical_and(Vx == 0., Vy == 0.))
    # additional fields if necessary
    if add_fields and fmt in [1, 3]:
        suppl_fields = []
        for i in np.arange(4, v_array.shape[0]):
            suppl_fields.append(np.transpose(np.array(v_array[i])))
    # Get and apply scale on values
    scale_i = atts['_SCALE_I']
    scale_i = scale_i.split("\n")
    unit_values = scale_i[1]
    scale_val = scale_i[0].split(' ')
    try:
        Vx *= int(scale_val[0])
        Vx += int(scale_val[1])
        Vy *= int(scale_val[0])
        Vy += int(scale_val[1])
    except ValueError:
        Vx *= float(scale_val[0])
        Vx += float(scale_val[1])
        Vy *= float(scale_val[0])
        Vy += float(scale_val[1])
    # Get and apply scale on X
    scale_x = atts['_SCALE_X']
    scale_x = scale_x.split("\n")
    unit_x = scale_x[1]
    scale_val = scale_x[0].split(' ')
    x_init = float(scale_val[1])
    dx = float(scale_val[0])*vectorGrid
    len_axe_x = Vx.shape[0]
    if dx < 0:
        axe_x = x_init + np.arange(len_axe_x - 1, -1, -1)*dx
        Vx = -Vx[::-1, :]
        Vy = Vy[::-1, :]
        mask = mask[::-1, :]
        if add_fields:
            for i in np.arange(len(suppl_fields)):
                suppl_fields[i] = suppl_fields[i][::-1, :]
    else:
        axe_x = x_init + np.arange(len_axe_x)*dx
    # Get and apply scale on Y
    scale_y = atts['_SCALE_Y']
    scale_y = scale_y.split("\n")
    unit_y = scale_y[1]
    scale_val = scale_y[0].split(' ')
    y_init = float(scale_val[1])
    dy = float(scale_val[0])*vectorGrid
    len_axe_y = Vx.shape[1]
    if dy < 0 or scale_y[1] == 'pixel':
        axe_y = y_init + np.arange(len_axe_y - 1, -1, -1)*dy
        Vx = Vx[:, ::-1]
        Vy = -Vy[:, ::-1]
        mask = mask[:, ::-1]
        if add_fields:
            for i in np.arange(len(suppl_fields)):
                suppl_fields[i] = suppl_fields[i][:, ::-1]
    else:
        axe_y = y_init + np.arange(len_axe_y)*dy
    # returning
    tmpvf = VectorField()
    tmpvf.import_from_arrays(axe_x, axe_y, Vx, Vy, mask=mask, unit_x=unit_x,
                             unit_y=unit_y, unit_values=unit_values)
    if not infos and not add_fields:
        return tmpvf
    res = ()
    res += (tmpvf,)
    if infos:
        res += (atts,)
    if add_fields:
        add_fields = []
        for i in np.arange(len(suppl_fields)):
            tmp_field = ScalarField()
            tmp_field.import_from_arrays(axe_x, axe_y, suppl_fields[i],
                                         unit_x=unit_x, unit_y=unit_y,
                                         unit_values='')
            add_fields.append(tmp_field)
        res += (add_fields,)
    return res


def import_from_VC7s(fieldspath, kind='TVF', fieldnumbers=None, incr=1,
                     add_fields=False, verbose=False):
    """
    Import velocity fields from .VC7 files.
    'fieldspath' should be a tuple of path to vc7 files.
    All vc7 file present in the folder are imported.

    Parameters
    ----------
    fieldspath : string or tuple of string
        If no '.vc7' are found directly under 'fieldspath', present folders are
        recursively serached for '.vc7' files.
    kind : string, optional
        Kind of object to create with VC7 files.
        (can be 'TVF' or 'SVF').
    fieldnumbers : 2x1 tuple of int
        Interval of fields to import, default is all.
    incr : integer
        Incrementation between fields to take. Default is 1, meaning all
        fields are taken.
    add_fields : boolean, optional
        If 'True', also return a tuple containing additional fields
        contained in the vc7 field (peak ratio, correlation value, ...).
    Verbose : bool, optional
        .
    """
    # check and adpat 'fieldspath'
    if isinstance(fieldspath, ARRAYTYPES):
        if not isinstance(fieldspath[0], STRINGTYPES):
            raise TypeError("'fieldspath' must be a string or a tuple of"
                            " string")
        paths = fieldspath
    elif isinstance(fieldspath, STRINGTYPES):
        fieldspath = check_path(fieldspath)
        paths = np.array([f for f in glob(path.join(fieldspath, '*'))
                          if path.splitext(f)[-1] in ['.vc7', '.VC7']])
        # if no file found, search recursively
        if len(paths) == 0:
            poss_paths = find_file_in_path(['.*.vc7', '.*.VC7'], fieldspath,
                                           ask=True)
            if len(poss_paths) == 0:
                raise ValueError()
            paths = np.array([f for f in glob(path.join(poss_paths[0], '*'))
                              if path.splitext(f)[-1] in ['.vc7', '.VC7']])
        # Sort path by numbers
        filenames = [path.basename(p) for p in paths]
        ind_sort = np.argsort(filenames)
        paths = paths[ind_sort]
    else:
        raise TypeError()
    # check and adapt 'fieldnumbers'
    if fieldnumbers is not None:
        if not isinstance(fieldnumbers, ARRAYTYPES):
            raise TypeError("'fieldnumbers' must be a 2x1 array")
        if not len(fieldnumbers) == 2:
            raise TypeError("'fieldnumbers' must be a 2x1 array")
        if not isinstance(fieldnumbers[0], int) \
                or not isinstance(fieldnumbers[1], int):
            raise TypeError("'fieldnumbers' must be an array of integers")
    else:
        fieldnumbers = [0, len(paths)]
    # check and adpat 'incr'
    if not isinstance(incr, int):
        raise TypeError("'incr' must be an integer")
    if incr <= 0:
        raise ValueError("'incr' must be positive")
    # Prepare containers
    if kind == 'TVF':
        fields = TemporalVectorFields()
    elif kind == 'SVF':
        fields = SpatialVectorFields()
    else:
        raise ValueError()
    # initialize counter
    start = fieldnumbers[0]
    end = fieldnumbers[1]
    nmb_files = int((end - start)/incr)
    pc = ProgressCounter(init_mess="Begin importation of {} VC7 files"
                         .format(nmb_files),
                         nmb_max=nmb_files,
                         name_things="VC7 files",
                         perc_interv=10)
    # loop on files
    t = 0.
    if add_fields:
        tmp_vf, add_fields = import_from_VC7(paths[0], add_fields=True)
        suppl_fields = [TemporalScalarFields() for field in add_fields]
    for i, p in enumerate(paths[start:end:incr]):
        if verbose:
            pc.print_progress()
        if add_fields:
            tmp_vf, infos, add_fields = import_from_VC7(p, infos=True,
                                                        add_fields=True)
            dt = infos['FrameDt0'].split()
            unit_time = make_unit(dt[1])
            dt = float(dt[0])
            t += dt*incr
            fields.add_field(tmp_vf, t, unit_time)
            for i, f in enumerate(add_fields):
                suppl_fields[i].add_field(f, t, unit_time)
        else:
            tmp_vf, infos = import_from_VC7(p, infos=True)
            try:
                dt = infos['FrameDt0'].split()
                unit_time = make_unit(dt[1])
                dt = float(dt[0])
            except KeyError:
                dt = 1.
                unit_time = make_unit("")
            t += dt*incr
            fields.add_field(tmp_vf, t, unit_time)
    # return
    if add_fields:
        return fields, suppl_fields
    else:
        return fields


def IM7_to_imt(im7_path, imt_path, kind='SF', compressed=True, **kwargs):
    """
    Transfome an IM7 (davis) file into a, imt exploitable file.

    Parameters
    ----------
    im7_path : path to file or directory
        Path to the IM7 file(s) , can be path to a single file or path to
        a directory contening multiples files.
    imt_path : path to file or directory
        Path where to save imt files, has to be the same type of path
        than 'im7_path' (path to file or path to directory)
    kind : string
        Kind of object to store (can be 'TSF' for TemporalScalarFields,
        'SSF' for SpatialScalarFields or 'SF' for multiple ScalarField)
    compressed : boolean, optional
        If 'True' (default), the file is compressed using gzip.
    kwargs : dict, optional
        Additional arguments for 'import_from_***()'.
    """
    # checking parameters
    im7_path = check_path(im7_path)
    imt_path = check_path(imt_path, newfile=True)
    # checking if file or directory
    if path.isdir(im7_path):
        if kind in ['SSF', 'TSF']:
            ST_SF = import_from_IM7s(im7_path, kind=kind, **kwargs)
            export_to_file(ST_SF, imt_path)
        elif kind in ['SF']:
            paths = glob(im7_path + "/*")
            for tmp_path in paths:
                name_ext = path.basename(tmp_path)
                name, ext = path.splitext(name_ext)
                if ext not in ['.im7', '.IM7']:
                    continue
                SF = import_from_IM7(tmp_path)
                export_to_file(SF, imt_path + "/{}".format(name),
                               compressed=compressed)
    elif path.isfile(im7_path):
        SF = import_from_IM7(im7_path, **kwargs)
        export_to_file(SF, imt_path)
    else:
        raise ValueError()


def VC7_to_imt(vc7_path, imt_path, kind='VF', compressed=True, **kwargs):
    """
    Transfome an VC7 (davis) file into a, imt exploitable file.

    Parameters
    ----------
    vc7_path : path to file or directory
        Path to the VC7 file(s) , can be path to a single file or path to
        a directory contening multiples files.
    imt_path : path to file
        Path where to save imt file.
    kind : string
        Kind of object to store (can be 'TVF' for TemporalVectorFields,
        'SVF' for SpatialVectorFields or 'VF' for multiple VectorField)
    compressed : boolean, optional
        If 'True' (default), the file is compressed using gzip.
    kwargs : dict, optional
        Additional arguments for 'import_from_***()'.
    """
    # checking parameters
    vc7_path = check_path(vc7_path)
    imt_path = check_path(imt_path)
    # checking if file or directory
    if path.isdir(vc7_path):
        if kind in ['SVF', 'TVF']:
            ST_VF = import_from_VC7s(vc7_path, kind=kind, **kwargs)
            export_to_file(ST_VF, imt_path)
        elif kind in ['VF']:
            paths = glob(vc7_path + "/*")
            for tmp_path in paths:
                name_ext = path.basename(tmp_path)
                name, ext = path.splitext(name_ext)
                if ext not in ['.vc7', '.VC7']:
                    continue
                VF = import_from_VC7(tmp_path)
                export_to_file(VF, imt_path + "/{}".format(name),
                               compressed=compressed)
    elif path.isfile(vc7_path):
        SF = import_from_IM7(vc7_path, **kwargs)
        export_to_file(SF, imt_path)
    else:
        raise ValueError()

def import_from_picture(filepath, axe_x=None, axe_y=None, unit_x='', unit_y='',
                        unit_values='', dtype=float):
    """
    Import a scalar field from a picture file.

    Parameters
    ----------
    filepath : string
        Path to the picture file.
    axe_x :
        .
    axe_y :
        .
    unit_x :
        .
    unit_y :
        .
    unit_values :
        .
    dtype :
        Type of the stored values (default to float)

    Returns
    -------
    tmp_sf :
        .
    """
    usable_ext = ['.png', '.PNG', '.jpg', '.JPG', '.jpeg', '.JPEG', '.bmp',
                  '.BMP']
    filepath = check_path(filepath)
    _, ext = path.splitext(filepath)
    if ext not in usable_ext:
        raise ValueError("I need the file to be an supported picture file"
                         "(not a {} file)".format(ext))
    # importing from file
    values = imageio.imread(filepath, as_gray=True).transpose()[:, ::-1]
    values = np.asarray(values, dtype=dtype)
    # set axis
    if axe_x is None:
        axe_x = np.arange(values.shape[0])
    else:
        if len(axe_x) != values.shape[0]:
            raise ValueError()
    if axe_y is None:
        axe_y = np.arange(values.shape[1])
    else:
        if len(axe_y) != values.shape[1]:
            raise ValueError()
    # create SF
    tmp_sf = ScalarField()
    tmp_sf.import_from_arrays(axe_x, axe_y, values, unit_x=unit_x,
                              unit_y=unit_y, unit_values=unit_values,
                              dtype=dtype)
    # return
    return tmp_sf

def import_from_pictures(filepath, axe_x=None, axe_y=None, unit_x='',
                         unit_y='', unit_values='', times=None,
                         unit_times='', dtype=float, fieldnumbers=None, incr=1,
                         verbose=False):
    """
    Import scalar fields from a bunch of picture files.

    Parameters
    ----------
    filepath : string
        regex matching the files.
    axe_x :
        .
    axe_y :
        .
    unit_x :
        .
    unit_y :
        .
    unit_values :
        .
    dtype :
        Type of the stored values (default to float)
    fieldnumbers: 2x1 array
        Interval of fields to import, default is all.
    incr : integer
        Increment (incr=2 will load only 1 picture over 2).

    Returns
    -------
    tmp_sf :
        .
    """
    # get paths
    # filepath = check_path(filepath)
    paths = glob(filepath)
    paths = sorted(paths)
    tmp_tsf = TemporalScalarFields()
    # check times
    if times is None:
        times = np.arange(len(paths))
    elif len(times) != len(paths):
        raise ValueError()
    # filter by field number
    if fieldnumbers is None:
        start = 0
        end = len(paths)
    else:
        if fieldnumbers[0] < 0:
            raise ValueError()
        if fieldnumbers[1] > len(paths):
            fieldnumbers[1] = len(paths)
        start = fieldnumbers[0]
        end = fieldnumbers[1]
    if verbose:
        pg = ProgressCounter(init_mess="Importing pictures",
                             nmb_max=int((end - start + 1)/incr),
                             name_things="pictures")
    # loop on paths
    for i in np.arange(start, end, incr):
        tmp_sf = import_from_picture(paths[i], axe_x=axe_x, axe_y=axe_y,
                                     unit_x=unit_x, unit_y=unit_y,
                                     unit_values=unit_values,
                                     dtype=dtype)
        tmp_tsf.add_field(tmp_sf, times[i], unit_times=unit_times)
        if verbose:
            pg.print_progress()
    # returning
    return tmp_tsf


def import_from_video(filepath, dx=1, dy=1, unit_x='',
                      unit_y='', unit_values='', dt=None,
                      unit_times='', dtype=float, frame_inds=None, incr=1,
                      verbose=False):
    """
    Import scalar fields from a video.

    Parameters
    ----------
    filepath : string
        regex matching the files.
    dx :
        .
    dy :
        .
    unit_x :
        .
    unit_y :
        .
    unit_values :
        .
    dt: number
        Time interval between two frames. If not specified, is estimated
        from the video.
    dtype :
        Type of the stored values (default to float)
    frame_inds: 2x1 array
        Interval of fields to import, default is all.
    incr : integer
        Increment (incr=2 will load only 1 picture over 2).

    Returns
    -------
    tmp_sf :
        .
    """
    # Chheck if cv2 is available
    if not IS_CV2:
        raise Exception('This feature needs opencv to be installed')
    # Check file
    filepath = check_path(filepath)
    paths = glob(filepath)
    paths = sorted(paths)
    # Open video stream
    vid = cv2.VideoCapture()
    vid.open(filepath)
    max_frame = int(vid.get(cv2.CAP_PROP_FRAME_COUNT))
    # Get framerate
    if dt is None:
        fps = float(vid.get(cv2.CAP_PROP_FPS))
        if fps == 0:
            fps = 1
        dt = 1/fps
    # filter by field number
    if frame_inds is None:
        frame_inds = [0, max_frame]
    if frame_inds[1] > max_frame:
        frame_inds[1] = max_frame
    if verbose:
        start = frame_inds[0]
        end = frame_inds[1]
        pg = ProgressCounter(init_mess="Importing video",
                             nmb_max=int((end - start + 1)/incr),
                             name_things="frames")
    # loop on paths
    tsf = TemporalScalarFields()
    t = 0
    for i in np.arange(0, frame_inds[1], 1):
        if i < frame_inds[0] or i % incr != 0:
            t += dt
            vid.grab()
            continue
        success, im = vid.read()
        if not success:
            if frame_inds[1] != np.inf:
                warnings.warn(f"Can't decode frame number {i}")
            break
        im = cv2.cvtColor(im, cv2.COLOR_RGB2GRAY)
        im = im.transpose()[:, ::-1]
        axe_x = np.arange(0, im.shape[0]*dx - 0.1*dx, dx)
        axe_y = np.arange(0, im.shape[1]*dy - 0.1*dy, dy)
        sf = ScalarField()
        sf.import_from_arrays(axe_x, axe_y, im, mask=False,
                              unit_x=unit_x, unit_y=unit_y,
                              dtype=dtype)
        tsf.add_field(sf, time=t, unit_times=unit_times, copy=False)
        t += dt
        if verbose:
            pg.print_progress()
    if verbose:
        pg.finish()
    # returning
    return tsf


def export_to_picture(SF, filepath):
    """
    Export a scalar field to a picture file.

    Parameters
    ----------
    SF :
        .
    filepath : string
        Path to the picture file.
    """
    filepath = check_path(filepath)
    values = SF.values[:, ::-1].transpose()
    imageio.imwrite(filepath, values)


def export_to_pictures(SFs, filepath):
    """
    Export scalar fields to a picture file.

    Parameters
    ----------
    SF :
        .
    filename : string
        Path to the picture file. Should include a name for the image
        (without the extension).
    """
    #check
    filepath = check_path(filepath, newfile=True)
    # get
    values = []
    if isinstance(SFs, ARRAYTYPES):
        for i in np.arange(len(SFs)):
            values.append(SFs[i].values)
    elif isinstance(SFs, (SpatialScalarFields, TemporalScalarFields)):
        for i in np.arange(len(SFs.fields)):
            values.append(SFs.fields[i].values[:, ::-1].transpose())
    # save
    for i, val in enumerate(values):
        imageio.imwrite("{}_{:0>5}.png".format(filepath, i), val)


def export_to_video(SFs, filepath, fps=24, colormap=None):
    """
    Export scalar fields to a video file.

    Parameters
    ----------
    SF :
        .
    filename : string
        Path to the video file.
    """
    #check
    filepath = check_path(filepath, newfile=True)
    # get
    values = []
    if isinstance(SFs, ARRAYTYPES):
        for i in np.arange(len(SFs)):
            values.append(SFs[i].values)
    elif isinstance(SFs, (SpatialScalarFields, TemporalScalarFields)):
        for i in np.arange(len(SFs.fields)):
            values.append(SFs.fields[i].values[:, ::-1].transpose())
    values = np.array(values)
    # normalize between 0 and 255
    maxi = np.max(values)
    mini = np.min(values)
    values = (values - mini)/(maxi - mini)
    # save as a video using opencv
    vid = cv2.VideoWriter(filename=filepath,
                          fourcc=cv2.VideoWriter_fourcc(*"MJPG"),
                          fps=fps,
                          frameSize=(values[0].shape[1],
                                     values[0].shape[0]))
    for val in values:
        if colormap is None:
            tmpval = cv2.cvtColor((val*255).astype('uint8'),
                                  cv2.COLOR_GRAY2RGB)
        else:
            tmpval = np.array(colormap(val)[:, :, 0:3]*255,
                              dtype=np.uint8)
            tmpval = tmpval[:, :, ::-1]
        # # TEMP
        # tmpval2 = cv2.cvtColor((val*255).astype('uint8'),
        #                        cv2.COLOR_GRAY2RGB)
        # tmpval[:, 0:int(tmpval.shape[1]/2), :] \
        #     = tmpval2[:, 0:int(tmpval.shape[1]/2), :]
        # # TEMP - End
        vid.write(tmpval)
    vid.release()

def import_profile_from_ascii(filepath, x_col=1, y_col=2,
                              unit_x=make_unit(""), unit_y=make_unit(""),
                              **kwargs):
        """
        Import a Profile object from an ascii file.

        Parameters
        ----------
        x_col, y_col : integer, optional
            Colonne numbers for the given variables (begining at 1).
        unit_x, unit_y : Unit objects, optional
            Unities for the given variables.
        **kwargs :
            Possibles additional parameters are the same as those used in the
            numpy function 'genfromtext()' :
            'delimiter' to specify the delimiter between colonnes.
            'skip_header' to specify the number of colonne to skip at file
                begining
        """
        # check
        filepath = check_path(filepath)
        # validating parameters
        if not isinstance(x_col, int) or not isinstance(y_col, int):
            raise TypeError("'x_col', 'y_col' must be integers")
        if x_col < 1 or y_col < 1:
            raise ValueError("Colonne number out of range")
        # 'names' deletion, if specified (dangereux pour la suite)
        if 'names' in kwargs:
            kwargs.pop('names')
        # extract data from file
        data = np.genfromtxt(filepath, **kwargs)
        # get axes
        x = data[:, x_col-1]
        y = data[:, y_col-1]
        prof = Profile(x, y, mask=False, unit_x=unit_x, unit_y=unit_y)
        return prof


def import_pts_from_ascii(filepath, x_col=1, y_col=2, v_col=None,
                          unit_x=make_unit(""), unit_y=make_unit(""),
                          unit_v=make_unit(""), **kwargs):
        """
        Import a Points object from an ascii file.

        Parameters
        ----------
        x_col, y_col, v_col : integer, optional
            Colonne numbers for the given variables
            (begining at 1).
        unit_x, unit_y, unit_v : Unit objects, optional
            Unities for the given variables.
        **kwargs :
            Possibles additional parameters are the same as those used in the
            numpy function 'genfromtext()' :
            'delimiter' to specify the delimiter between colonnes.
            'skip_header' to specify the number of colonne to skip at file
                begining
            ...
        """
        # check
        filepath = check_path(filepath)
        # validating parameters
        if v_col is None:
            v_col = 0
        if not isinstance(x_col, int) or not isinstance(y_col, int)\
                or not isinstance(v_col, int):
            raise TypeError("'x_col', 'y_col' and 'v_col' must be integers")
        if x_col < 1 or y_col < 1:
            raise ValueError("Colonne number out of range")
        # 'names' deletion, if specified (dangereux pour la suite)
        if 'names' in kwargs:
            kwargs.pop('names')
        # extract data from file
        data = np.genfromtxt(filepath, **kwargs)
        # get axes
        x = data[:, x_col-1]
        y = data[:, y_col-1]
        if v_col != 0:
            v = data[:, v_col-1]
        else:
            v = None
        return Points(list(zip(x, y)), v, unit_x, unit_y, unit_v)


def import_sf_from_ascii(filepath, x_col=1, y_col=2, vx_col=3,
                         unit_x=make_unit(""),
                         unit_y=make_unit(""),
                         unit_values=make_unit(""), **kwargs):
    """
    Import a scalarfield from an ascii file.

    Parameters
    ----------
    x_col, y_col, vx_col: integer, optional
        Colonne numbers for the given variables
        (begining at 1).
    unit_x, unit_y, unit_v : Unit objects, optional
        Unities for the given variables.
    **kwargs :
        Possibles additional parameters are the same as those used in the
        numpy function 'genfromtext()' :
        'delimiter' to specify the delimiter between colonnes.
        'skip_header' to specify the number of colonne to skip at file
            begining
        ...
    """
    # check
    filepath = check_path(filepath)
    # validating parameters
    if not isinstance(x_col, int) or not isinstance(y_col, int)\
            or not isinstance(vx_col, int):
        raise TypeError("'x_col', 'y_col', 'vx_col' and 'vy_col' must "
                        "be integers")
    if x_col < 1 or y_col < 1 or vx_col < 1:
        raise ValueError("Colonne number out of range")
    # 'names' deletion, if specified (dangereux pour la suite)
    if 'names' in kwargs:
        kwargs.pop('names')
    # extract data from file
    data = np.genfromtxt(filepath, **kwargs)
    # get axes
    x = data[:, x_col-1]
    x_org = np.sort(np.unique(x))
    y = data[:, y_col-1]
    y_org = np.sort(np.unique(y))
    vx = data[:, vx_col-1]

    # check if structured or not
    X1 = x.reshape(len(x_org), len(y_org))
    X2 = x.reshape(len(y_org), len(x_org))
    if np.allclose(np.mean(X1, axis=1), x_org):
        vx_org = vx.reshape(len(x_org), len(y_org))
        mask = np.isnan(vx_org)
    elif np.allclose(np.mean(X2, axis=0), x_org):
        vx_org = vx.reshape(len(y_org), len(x_org)).transpose()
        vx_org = np.fliplr(vx_org)
        mask = np.isnan(vx_org)
    else:
        # Masking all the initial fields (to handle missing values)
        vx_org = np.zeros((y_org.shape[0], x_org.shape[0]))
        vx_org_mask = np.ones(vx_org.shape)
        vx_org = np.ma.masked_array(vx_org, vx_org_mask)
        #loop on all 'v' values
        x_ind = 0
        y_ind = 0
        for i in np.arange(vx.shape[0]):
            x_tmp = x[i]
            y_tmp = y[i]
            vx_tmp = vx[i]
            #find x index
            if x_org[x_ind] != x_tmp:
                x_ind = np.where(x_tmp == x_org)[0][0]
            #find y index
            if y_org[y_ind] != y_tmp:
                y_ind = np.where(y_tmp == y_org)[0][0]
            #put the value at its place
            vx_org[y_ind, x_ind] = vx_tmp
        # Treating 'nan' values
        mask = np.logical_or(vx_org.mask, np.isnan(vx_org.data))

    #store field in attributes
    tmpsf = ScalarField()
    tmpsf.import_from_arrays(x_org, y_org, vx_org, mask=mask, unit_x=unit_x,
                             unit_y=unit_y, unit_values=unit_values)
    return tmpsf


def import_vf_from_ascii(filepath, x_col=1, y_col=2, vx_col=3,
                         vy_col=4, unit_x=make_unit(""),
                         unit_y=make_unit(""),
                         unit_values=make_unit(""), **kwargs):
    """
    Import a vectorfield from an ascii file.

    Parameters
    ----------
    x_col, y_col, vx_col, vy_col : integer, optional
        Colonne numbers for the given variables
        (begining at 1).
    unit_x, unit_y, unit_v : Unit objects, optional
        Unities for the given variables.
    **kwargs :
        Possibles additional parameters are the same as those used in the
        numpy function 'genfromtext()' :
        'delimiter' to specify the delimiter between colonnes.
        'skip_header' to specify the number of colonne to skip at file
            begining
        ...
    """
    # check
    filepath = check_path(filepath)
    # validating parameters
    if not isinstance(x_col, int) or not isinstance(y_col, int)\
            or not isinstance(vx_col, int):
        raise TypeError("'x_col', 'y_col', 'vx_col' and 'vy_col' must "
                        "be integers")
    if x_col < 1 or y_col < 1 or vx_col < 1:
        raise ValueError("Colonne number out of range")
    if vy_col is not None:
        if not isinstance(vy_col, int):
            raise TypeError("'x_col', 'y_col', 'vx_col' and 'vy_col' must "
                            "be integers")
        if vy_col < 1:
            raise ValueError("Colonne number out of range")
    # 'names' deletion, if specified (dangereux pour la suite)
    if 'names' in kwargs:
        kwargs.pop('names')
    # extract data from file
    data = np.genfromtxt(filepath, **kwargs)
    # get axes
    x = data[:, x_col-1]
    x_org = np.unique(x)
    y = data[:, y_col-1]
    y_org = np.unique(y)
    vx = data[:, vx_col-1]
    vy = data[:, vy_col-1]
    # Masking all the initial fields (to handle missing values)
    vx_org = np.zeros((x_org.shape[0], y_org.shape[0]))
    vx_org_mask = np.ones(vx_org.shape)
    vx_org = np.ma.masked_array(vx_org, vx_org_mask)
    vy_org = np.zeros((x_org.shape[0], y_org.shape[0]))
    vy_org_mask = np.ones(vy_org.shape)
    vy_org = np.ma.masked_array(vy_org, vy_org_mask)
    #loop on all 'v' values
    for i in np.arange(vx.shape[0]):
        x_tmp = x[i]
        y_tmp = y[i]
        vx_tmp = vx[i]
        vy_tmp = vy[i]
        #find x index
        for j in np.arange(x_org.shape[0]):
            if x_org[j] == x_tmp:
                x_ind = j
        #find y index
        for j in np.arange(y_org.shape[0]):
            if y_org[j] == y_tmp:
                y_ind = j
        #put the value at its place
        vx_org[x_ind, y_ind] = vx_tmp
        vy_org[x_ind, y_ind] = vy_tmp
    # Treating 'nan' values
    vx_org.mask = np.logical_or(vx_org.mask, np.isnan(vx_org.data))
    vy_org.mask = np.logical_or(vy_org.mask, np.isnan(vy_org.data))
    #store field in attributes
    tmpvf = VectorField()
    tmpvf.import_from_arrays(x_org, y_org, vx_org, vy_org,
                             mask=vx_org.mask,
                             unit_x=unit_x, unit_y=unit_y,
                             unit_values=unit_values)
    return tmpvf


def import_vfs_from_ascii(filepath, kind='TVF', incr=1, interval=None,
                          x_col=1, y_col=2, vx_col=3,
                          vy_col=4, unit_x=make_unit(""),
                          unit_y=make_unit(""),
                          unit_values=make_unit(""), times=[],
                          unit_time=make_unit(''), **kwargs):
    """
    Import velocityfields from an ascii file.

    Parameters
    ----------
    filepath : string
        Pathname pattern to the ascii files.
    incr : integer, optional
        Increment value between two fields taken.
    interval : 2x1 array, optional
        Interval in which take fields.
    x_col, y_col, vx_col, vy_col : integer, optional
        Colonne numbers for the given variables (begining at 1).
    unit_x, unit_y, unit_v : Unit objects, optional
        Unities for the given variables.
    times : array of number, optional
        Times of the instantaneous fields.
    unit_time : Unit object, optional
        Time unit, 'second' by default.
    **kwargs :
        Possibles additional parameters are the same as those used in the
        numpy function 'genfromtext()' :
        'delimiter' to specify the delimiter between colonnes.
        'skip_header' to specify the number of colonne to skip at file
        begining

    Note
    ----
    txt files are taken in alpha-numerical order
    ('file2.txt' is taken before 'file20.txt').
    So you should name your files properly.

    """
    if not isinstance(incr, int):
        raise TypeError("'incr' must be an integer")
    if incr < 1:
        raise ValueError("'incr' must be superior to 1")
    if interval is not None:
        if not isinstance(interval, ARRAYTYPES):
            raise TypeError("'interval' must be an array")
        if not len(interval) == 2:
            raise ValueError("'interval' must be a 2x1 array")
        if interval[0] > interval[1]:
            interval = [interval[1], interval[0]]
    filepath = check_path(filepath)
    paths = glob.glob(filepath)
    if interval is None:
        interval = [0, len(paths)-1]
    if interval[0] < 0 or interval[1] > len(paths):
        raise ValueError("'interval' is out of bounds")
    if times == []:
        times = np.arange(len(paths))
    if len(paths) != len(times):
        raise ValueError("Not enough values in 'times'")
    ref_path_len = len(paths[0])
    if kind == 'TVF':
        fields = TemporalVectorFields()
    elif kind == 'SVF':
        fields = SpatialVectorFields()
    else:
        raise ValueError()
    for i in np.arange(interval[0], interval[1] + 1, incr):
        path = paths[i]
        if len(path) != ref_path_len:
            raise Warning("You should check your files names,"
                          "i may have taken them in the wrong order.")
        tmp_vf = VectorField()
        tmp_vf.import_from_ascii(path, x_col, y_col, vx_col, vy_col,
                                 unit_x, unit_y, unit_values, times[i],
                                 unit_time, **kwargs)
        fields.add_field(tmp_vf)
    return fields


def export_to_ascii(obj, filepath):
    """
    """
    # check
    filepath = check_path(filepath, newfile=True)
    # obj type
    if isinstance(obj, VectorField):
        # open file
        f = open(filepath, 'w')
        # write header
        header = "# X {}\tY {}\tVx {}\tVy {}\n"\
                 .format(obj.unit_x.strUnit(),
                         obj.unit_y.strUnit(),
                         obj.unit_values.strUnit(),
                         obj.unit_values.strUnit())
        f.write(header)
        # write data
        for i, x in enumerate(obj.axe_x):
            for j, y in enumerate(obj.axe_y):
                f.write("{}\t{}\t{}\t{}\n".format(x, y, obj.comp_x[i, j],
                                                  obj.comp_y[i, j]))
        f.close()
    elif isinstance(obj, ScalarField):
        # open file
        f = open(filepath, 'w')
        # write header
        header = "# X {}\tY {}\tValue {}\n"\
                 .format(obj.unit_x.strUnit(),
                         obj.unit_y.strUnit(),
                         obj.unit_values.strUnit())
        f.write(header)
        # write data
        for i, x in enumerate(obj.axe_x):
            for j, y in enumerate(obj.axe_y):
                f.write("{}\t{}\t{}\n".format(x, y, obj.values[i, j]))
        f.close()
    elif isinstance(obj, Profile):
        # open file
        f = open(filepath, 'w')
        # write header
        header = "# X {}\tY {}\n"\
                 .format(obj.unit_x.strUnit(),
                         obj.unit_y.strUnit())
        f.write(header)
        # write data
        for x, y in zip(obj.x, obj.y):
            f.write("{}\t{}\n".format(x, y))
        f.close()
    elif isinstance(obj, Points):
        # open file
        f = open(filepath, 'w')
        if len(obj.v) == 0:
            # write header
            header = "# X {}\tY {}\n"\
                     .format(obj.unit_x.strUnit(),
                             obj.unit_y.strUnit())
            f.write(header)
            # write data
            for x, y in zip(obj.xy[:, 0], obj.xy[:, 1]):
                f.write("{}\t{}\n".format(x, y))
        else:
            # write header
            header = "# X {}\tY {}\tV {}\n"\
                     .format(obj.unit_x.strUnit(),
                             obj.unit_y.strUnit(),
                             obj.unit_v.strUnit())
            f.write(header)
            # write data
            for x, y, v in zip(obj.xy[:, 0], obj.xy[:, 1], obj.v):
                f.write("{}\t{}\t{}\n".format(x, y, v))
        f.close()
    elif isinstance(obj, CritPoints):
        # get info from path
        basename = os.path.splitext(filepath)[0]
        try:
            os.mkdir(basename)
        except FileExistsError:
            pass
        t_unit = obj.unit_time.strUnit()
        if t_unit == "[]":
            t_unit = ""
        for typ in obj.cp_types:
            pts = obj.__getattribute__(typ)
            for t, pt in zip(obj.times, pts):
                export_to_ascii(pt, "{}/{}_t={}{}.txt".format(basename,
                                                              typ, t, t_unit))
    else:
        raise TypeError()



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import numpy as np


def get_sol():
    # Computed with 'compute_coefs()'
    s00 = '-dx*(-Vx_1*Vy_2*dy - Vx_1*Vy_3*dy + Vx_1*Vy_4*dy + Vx_2*Vy_1*dy + Vx_3*Vy_1*dy - Vx_4*Vy_1*dy + (dy*(2*Vx_1*Vy_2 - Vx_1*Vy_4 - 2*Vx_2*Vy_1 + Vx_2*Vy_3 - Vx_3*Vy_2 + Vx_4*Vy_1)/(2*(Vx_1*Vy_2 - Vx_1*Vy_4 - Vx_2*Vy_1 + Vx_2*Vy_3 - Vx_3*Vy_2 + Vx_3*Vy_4 + Vx_4*Vy_1 - Vx_4*Vy_3)) - sqrt(dy**2*(Vx_1**2*Vy_4**2 - 2*Vx_1*Vx_2*Vy_3*Vy_4 - 2*Vx_1*Vx_3*Vy_2*Vy_4 - 2*Vx_1*Vx_4*Vy_1*Vy_4 + 4*Vx_1*Vx_4*Vy_2*Vy_3 + Vx_2**2*Vy_3**2 + 4*Vx_2*Vx_3*Vy_1*Vy_4 - 2*Vx_2*Vx_3*Vy_2*Vy_3 - 2*Vx_2*Vx_4*Vy_1*Vy_3 + Vx_3**2*Vy_2**2 - 2*Vx_3*Vx_4*Vy_1*Vy_2 + Vx_4**2*Vy_1**2))/(2*Vx_1*Vy_2 - 2*Vx_1*Vy_4 - 2*Vx_2*Vy_1 + 2*Vx_2*Vy_3 - 2*Vx_3*Vy_2 + 2*Vx_3*Vy_4 + 2*Vx_4*Vy_1 - 2*Vx_4*Vy_3))*(Vx_1*Vy_2 - Vx_1*Vy_4 - Vx_2*Vy_1 + Vx_2*Vy_3 - Vx_3*Vy_2 + Vx_3*Vy_4 + Vx_4*Vy_1 - Vx_4*Vy_3))/(dy*(Vx_1*Vy_3 - Vx_1*Vy_4 - Vx_2*Vy_3 + Vx_2*Vy_4 - Vx_3*Vy_1 + Vx_3*Vy_2 + Vx_4*Vy_1 - Vx_4*Vy_2))'
    s01 = 'dy*(2*Vx_1*Vy_2 - Vx_1*Vy_4 - 2*Vx_2*Vy_1 + Vx_2*Vy_3 - Vx_3*Vy_2 + Vx_4*Vy_1)/(2*(Vx_1*Vy_2 - Vx_1*Vy_4 - Vx_2*Vy_1 + Vx_2*Vy_3 - Vx_3*Vy_2 + Vx_3*Vy_4 + Vx_4*Vy_1 - Vx_4*Vy_3)) - sqrt(dy**2*(Vx_1**2*Vy_4**2 - 2*Vx_1*Vx_2*Vy_3*Vy_4 - 2*Vx_1*Vx_3*Vy_2*Vy_4 - 2*Vx_1*Vx_4*Vy_1*Vy_4 + 4*Vx_1*Vx_4*Vy_2*Vy_3 + Vx_2**2*Vy_3**2 + 4*Vx_2*Vx_3*Vy_1*Vy_4 - 2*Vx_2*Vx_3*Vy_2*Vy_3 - 2*Vx_2*Vx_4*Vy_1*Vy_3 + Vx_3**2*Vy_2**2 - 2*Vx_3*Vx_4*Vy_1*Vy_2 + Vx_4**2*Vy_1**2))/(2*Vx_1*Vy_2 - 2*Vx_1*Vy_4 - 2*Vx_2*Vy_1 + 2*Vx_2*Vy_3 - 2*Vx_3*Vy_2 + 2*Vx_3*Vy_4 + 2*Vx_4*Vy_1 - 2*Vx_4*Vy_3)'
    s10 = '-dx*(-Vx_1*Vy_2*dy - Vx_1*Vy_3*dy + Vx_1*Vy_4*dy + Vx_2*Vy_1*dy + Vx_3*Vy_1*dy - Vx_4*Vy_1*dy + (dy*(2*Vx_1*Vy_2 - Vx_1*Vy_4 - 2*Vx_2*Vy_1 + Vx_2*Vy_3 - Vx_3*Vy_2 + Vx_4*Vy_1)/(2*(Vx_1*Vy_2 - Vx_1*Vy_4 - Vx_2*Vy_1 + Vx_2*Vy_3 - Vx_3*Vy_2 + Vx_3*Vy_4 + Vx_4*Vy_1 - Vx_4*Vy_3)) + sqrt(dy**2*(Vx_1**2*Vy_4**2 - 2*Vx_1*Vx_2*Vy_3*Vy_4 - 2*Vx_1*Vx_3*Vy_2*Vy_4 - 2*Vx_1*Vx_4*Vy_1*Vy_4 + 4*Vx_1*Vx_4*Vy_2*Vy_3 + Vx_2**2*Vy_3**2 + 4*Vx_2*Vx_3*Vy_1*Vy_4 - 2*Vx_2*Vx_3*Vy_2*Vy_3 - 2*Vx_2*Vx_4*Vy_1*Vy_3 + Vx_3**2*Vy_2**2 - 2*Vx_3*Vx_4*Vy_1*Vy_2 + Vx_4**2*Vy_1**2))/(2*Vx_1*Vy_2 - 2*Vx_1*Vy_4 - 2*Vx_2*Vy_1 + 2*Vx_2*Vy_3 - 2*Vx_3*Vy_2 + 2*Vx_3*Vy_4 + 2*Vx_4*Vy_1 - 2*Vx_4*Vy_3))*(Vx_1*Vy_2 - Vx_1*Vy_4 - Vx_2*Vy_1 + Vx_2*Vy_3 - Vx_3*Vy_2 + Vx_3*Vy_4 + Vx_4*Vy_1 - Vx_4*Vy_3))/(dy*(Vx_1*Vy_3 - Vx_1*Vy_4 - Vx_2*Vy_3 + Vx_2*Vy_4 - Vx_3*Vy_1 + Vx_3*Vy_2 + Vx_4*Vy_1 - Vx_4*Vy_2))'
    s11 = 'dy*(2*Vx_1*Vy_2 - Vx_1*Vy_4 - 2*Vx_2*Vy_1 + Vx_2*Vy_3 - Vx_3*Vy_2 + Vx_4*Vy_1)/(2*(Vx_1*Vy_2 - Vx_1*Vy_4 - Vx_2*Vy_1 + Vx_2*Vy_3 - Vx_3*Vy_2 + Vx_3*Vy_4 + Vx_4*Vy_1 - Vx_4*Vy_3)) + sqrt(dy**2*(Vx_1**2*Vy_4**2 - 2*Vx_1*Vx_2*Vy_3*Vy_4 - 2*Vx_1*Vx_3*Vy_2*Vy_4 - 2*Vx_1*Vx_4*Vy_1*Vy_4 + 4*Vx_1*Vx_4*Vy_2*Vy_3 + Vx_2**2*Vy_3**2 + 4*Vx_2*Vx_3*Vy_1*Vy_4 - 2*Vx_2*Vx_3*Vy_2*Vy_3 - 2*Vx_2*Vx_4*Vy_1*Vy_3 + Vx_3**2*Vy_2**2 - 2*Vx_3*Vx_4*Vy_1*Vy_2 + Vx_4**2*Vy_1**2))/(2*Vx_1*Vy_2 - 2*Vx_1*Vy_4 - 2*Vx_2*Vy_1 + 2*Vx_2*Vy_3 - 2*Vx_3*Vy_2 + 2*Vx_3*Vy_4 + 2*Vx_4*Vy_1 - 2*Vx_4*Vy_3)'
    sol = np.array([[s00, s01], [s10, s11]])
    return sol


def compute_coefs():
    try:
        import sympy
    except ImportError:
        raise Exception('You will need \'sympy\' to use this function')
        # prepare the analytical solution
    a1, b1, c1, d1, x, y = sympy.symbols('a1, b1, c1, d1, x, y')
    def funct(params, x, y, Vx):
        a1, b1, c1, d1 = tuple(params)
        return sympy.Eq(a1*x + b1*y + c1*x*y + d1, Vx)
    def funct1(dx, dy, Vx):
        return sympy.Eq(d1, Vx)
    def funct2(dx, dy, Vx):
        return sympy.Eq(a1*dx + d1, Vx)
    def funct3(dx, dy, Vx):
        return sympy.Eq(b1*dy + d1, Vx)
    def funct4(dx, dy, Vx):
        return sympy.Eq(a1*dx + b1*dy + c1*dx*dy + d1, Vx)
    # get the analytical solution for arbitrary Vx and Vy
    Vx_1, Vx_2, Vx_3, Vx_4 = sympy.symbols('Vx_1, Vx_2, Vx_3, Vx_4')
    Vy_1, Vy_2, Vy_3, Vy_4 = sympy.symbols('Vy_1, Vy_2, Vy_3, Vy_4')
    dx, dy = sympy.symbols('dx, dy')
    bl11 = funct1(dx, dy, Vx_1)
    bl12 = funct2(dx, dy, Vx_2)
    bl13 = funct3(dx, dy, Vx_3)
    bl14 = funct4(dx, dy, Vx_4)
    sol1 = sympy.solve([bl11, bl12, bl13, bl14], [a1, b1, c1, d1])
    params = [sol1[a1], sol1[b1], sol1[c1], sol1[d1]]
    eq1 = funct(params, x, y, 0.)
    eq2 = eq1.subs(Vx_1, Vy_1)
    eq2 = eq2.subs(Vx_2, Vy_2)
    eq2 = eq2.subs(Vx_3, Vy_3)
    eq2 = eq2.subs(Vx_4, Vy_4)
    sol = sympy.solve([eq1, eq2], [x, y])
    sol = np.array(sol)
    if sol.ndim == 1:
        sol = np.array([sol])
    for i in np.arange(sol.shape[0]):
        for j in np.arange(sol.shape[1]):
            sol[i, j] = sol[i, j].__repr__()
    return sol



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import copy
import warnings

import matplotlib.pyplot as plt
import numpy as np
from scipy import ndimage

from .. import plotlib as pplt
import unum
from . import field, scalarfield as sf
from ..utils import make_unit
from ..utils.types import ARRAYTYPES, NUMBERTYPES, STRINGTYPES


class VectorField(field.Field):
    """
    Class representing a vector field (2D field, with two components on each
    point).
    """

    def __init__(self):
        super(VectorField, self).__init__()
        self.__comp_x = np.array([], dtype=float)
        self.__comp_y = np.array([], dtype=float)
        self.__mask = np.array([], dtype=bool)
        self.__unit_values = make_unit('')
        self._values_dtype = None

    def __neg__(self):
        tmpvf = self.copy()
        tmpvf.comp_x = -tmpvf.comp_x
        tmpvf.comp_y = -tmpvf.comp_y
        return tmpvf

    def __add__(self, other):
        if isinstance(other, VectorField):
            # test unities system
            try:
                self.unit_values + other.unit_values
                self.unit_x + other.unit_x
                self.unit_y + other.unit_y
            except:
                raise ValueError("I think these units don't match, fox")
            # identical shape and axis
            if np.all(self.axe_x == other.axe_x) and \
                    np.all(self.axe_y == other.axe_y):
                tmpvf = self.copy()
                fact = (other.unit_values/self.unit_values).asNumber()
                tmpvf.comp_x = self.comp_x + other.comp_x*fact
                tmpvf.comp_y = self.comp_y + other.comp_y*fact
                tmpvf.mask = np.logical_or(self.mask, other.mask)
                return tmpvf
            # different shape, partially same axis
            else:
                # getting shared points
                new_ind_x = np.array([np.any(np.abs(val - other.axe_x) <
                                      np.abs(val)*1e-4)
                                      for val in self.axe_x])
                new_ind_y = np.array([np.any(np.abs(val - other.axe_y) <
                                      np.abs(val)*1e-4)
                                      for val in self.axe_y])
                new_ind_xo = np.array([np.any(np.abs(val - self.axe_x) <
                                       np.abs(val)*1e-4)
                                       for val in other.axe_x])
                new_ind_yo = np.array([np.any(np.abs(val - self.axe_y) <
                                       np.abs(val)*1e-4)
                                       for val in other.axe_y])
                if not np.any(new_ind_x) or not np.any(new_ind_y):
                    raise ValueError("Incompatible shapes")
                new_ind_Y, new_ind_X = np.meshgrid(new_ind_y, new_ind_x)
                new_ind_value = np.logical_and(new_ind_X, new_ind_Y)
                new_ind_Yo, new_ind_Xo = np.meshgrid(new_ind_yo, new_ind_xo)
                new_ind_valueo = np.logical_and(new_ind_Xo, new_ind_Yo)
                # getting new axis and values
                new_axe_x = self.axe_x[new_ind_x]
                new_axe_y = self.axe_y[new_ind_y]
                fact = other.unit_values/self.unit_values
                new_comp_x = (self.comp_x[new_ind_value] +
                              other.comp_x[new_ind_valueo] *
                              fact.asNumber())
                new_comp_y = (self.comp_y[new_ind_value] +
                              other.comp_y[new_ind_valueo] *
                              fact.asNumber())
                new_comp_x = new_comp_x.reshape((len(new_axe_x),
                                                 len(new_axe_y)))
                new_comp_y = new_comp_y.reshape((len(new_axe_x),
                                                 len(new_axe_y)))
                new_mask = np.logical_or(self.mask[new_ind_value],
                                         other.mask[new_ind_valueo])
                new_mask = new_mask.reshape((len(new_axe_x), len(new_axe_y)))
                # creating vf
                tmpvf = VectorField()
                tmpvf.import_from_arrays(new_axe_x, new_axe_y, new_comp_x,
                                         new_comp_y,
                                         mask=new_mask, unit_x=self.unit_x,
                                         unit_y=self.unit_y,
                                         unit_values=self.unit_values)
                return tmpvf
        elif isinstance(other, ARRAYTYPES):
            other = np.array(other, subok=True)
            # Same shape
            if other.shape == self.shape:
                tmpvf = self.copy()
                tmpvf.comp_x = self.comp_x + other
                tmpvf.comp_y = self.comp_y + other
                tmpvf.mask = self.mask
                return tmpvf
            # signle vector
            elif other.shape == (2,):
                tmpvf = self.copy()
                tmpvf.comp_x += other[0]
                tmpvf.comp_y += other[1]
                tmpvf.mask = self.mask
                return tmpvf
            else:
                raise ValueError()
        elif isinstance(other, unum.Unum):
            tmpvf = self.copy()
            fact = (other / self.unit_values).asNumber()
            tmpvf.comp_x = self.comp_x + fact
            tmpvf.comp_y = self.comp_y + fact
            tmpvf.mask = self.mask
            return tmpvf
        elif isinstance(other, NUMBERTYPES):
            tmpvf = self.copy()
            tmpvf.comp_x = self.comp_x + other
            tmpvf.comp_y = self.comp_y + other
            tmpvf.mask = self.mask
            return tmpvf
        else:
            raise TypeError("You can only add a velocity field "
                            "with others velocity fields")

    __radd__ = __add__

    def __sub__(self, other):
        other_tmp = other.__neg__()
        tmpvf = self.__add__(other_tmp)
        return tmpvf

    def __rsub__(self, other):
        return other + self.__neg__()

    def __truediv__(self, other):
        if isinstance(other, ARRAYTYPES):
            other = np.array(other, subok=True)
            if other.shape != self.shape:
                raise ValueError()
            tmpvf = self.copy()
            tmpvf.comp_x = self.comp_x / other
            tmpvf.comp_y = self.comp_y / other
            tmpvf.mask = np.logical_or(self.mask, other == 0)
            return tmpvf
        elif isinstance(other, unum.Unum):
            tmpvf = self.copy()
            new_unit = tmpvf.unit_values/other
            scale = new_unit.asNumber()
            new_unit /= scale
            tmpvf.unit_values = new_unit
            tmpvf.comp_x *= scale
            tmpvf.comp_y *= scale
            tmpvf.mask = self.mask
            return tmpvf
        elif isinstance(other, NUMBERTYPES):
            tmpvf = self.copy()
            tmpvf.comp_x /= other
            tmpvf.comp_y /= other
            tmpvf.mask = self.mask
            return tmpvf
        else:
            raise TypeError("You can only divide a vector field "
                            "by numbers")

    __div__ = __truediv__

    def __rtruediv__(self, other):
        return other * self**(-1)

    __rdiv__ = __rtruediv__

    def __mul__(self, other):
        if isinstance(other, ARRAYTYPES):
            other = np.array(other, subok=True)
            if other.shape != self.shape:
                raise ValueError()
            tmpvf = self.copy()
            tmpvf.comp_x = self.comp_x * other
            tmpvf.comp_y = self.comp_y * other
            tmpvf.mask = self.mask
            return tmpvf
        elif isinstance(other, unum.Unum):
            tmpvf = self.copy()
            new_unit = tmpvf.unit_values*other
            scale = new_unit.asNumber()
            new_unit /= scale
            tmpvf.unit_values = new_unit
            tmpvf.comp_x *= scale
            tmpvf.comp_y *= scale
            tmpvf.mask = self.mask
            return tmpvf
        elif isinstance(other, NUMBERTYPES):
            tmpvf = self.copy()
            tmpvf.comp_x *= other
            tmpvf.comp_y *= other
            tmpvf.mask = self.mask
            return tmpvf
        elif isinstance(other, sf.ScalarField):
            if other.shape != self.shape:
                raise ValueError()
            tmpvf = self.copy()
            tmpvf.comp_x *= other.values
            tmpvf.comp_y *= other.values
            tmpvf.mask = np.logical_or(other.mask, self.mask)
            return tmpvf
        else:
            raise TypeError("You cannot multiply Vectorfields object "
                            "with {} objects".format(type(other)))

    __rmul__ = __mul__

    def __pow__(self, number):
        if not isinstance(number, NUMBERTYPES):
            raise TypeError("You only can use a number for the power "
                            "on a Vectorfield")
        tmpvf = self.copy()
        tmpvf.comp_x = np.power(tmpvf.comp_x, number)
        tmpvf.comp_y = np.power(tmpvf.comp_y, number)
        tmpvf.unit_values = np.power(tmpvf.unit_values, number)
        return tmpvf

    def __abs__(self):
        tmpvf = self.copy()
        tmpvf.comp_x = np.abs(tmpvf.comp_x)
        tmpvf.comp_y = np.abs(tmpvf.comp_y)
        return tmpvf

    def __iter__(self):
        mask = self.mask
        datax = self.comp_x
        datay = self.comp_y
        for ij, xy in field.Field.__iter__(self):
            i = ij[0]
            j = ij[1]
            if not mask[i, j]:
                yield ij, xy, [datax[i, j], datay[i, j]]

    def __eq__(self, obj):
        if not isinstance(obj, VectorField):
            return False
        if not self.shape == obj.shape:
            return False
        if not np.all(self.axe_x == obj.axe_x):
            return False
        if not np.all(self.axe_y == obj.axe_y):
            return False
        if not np.all(self.mask == obj.mask):
            return False
        if not np.all(self.comp_x[~self.mask] == obj.comp_x[~self.mask]):
            return False
        if not np.all(self.comp_y[~self.mask] == obj.comp_y[~self.mask]):
            return False
        if not self.unit_x == obj.unit_x:
            return False
        if not self.unit_y == obj.unit_y:
            return False
        if not self.unit_values == obj.unit_values:
            return False
        return True

    @property
    def comp_x(self):
        return self.__comp_x

    @comp_x.setter
    def comp_x(self, new_comp_x):
        if not isinstance(new_comp_x, ARRAYTYPES):
            raise TypeError()
        new_comp_x = np.array(new_comp_x)
        if not new_comp_x.shape == self.shape:
            raise ValueError("'comp_x' must be coherent with axis system")
        # storing dat
        self.__comp_x = new_comp_x

    @comp_x.deleter
    def comp_x(self):
        raise Exception("Nope, can't do that")

    @property
    def comp_x_as_sf(self):
        tmp_sf = sf.ScalarField()
        tmp_sf.import_from_arrays(self.axe_x, self.axe_y, self.comp_x,
                                  mask=self.mask, unit_x=self.unit_x,
                                  unit_y=self.unit_y,
                                  unit_values=self.unit_values)
        return tmp_sf

    @property
    def comp_y(self):
        return self.__comp_y

    @comp_y.setter
    def comp_y(self, new_comp_y):
        if not isinstance(new_comp_y, ARRAYTYPES):
            raise TypeError()
        new_comp_y = np.array(new_comp_y)
        if not new_comp_y.shape == self.shape:
            raise ValueError()
        # storing data
        self.__comp_y = new_comp_y

    @comp_y.deleter
    def comp_y(self):
        raise Exception("Nope, can't do that")

    @property
    def comp_y_as_sf(self):
        tmp_sf = sf.ScalarField()
        tmp_sf.import_from_arrays(self.axe_x, self.axe_y, self.comp_y,
                                  mask=self.mask, unit_x=self.unit_x,
                                  unit_y=self.unit_y,
                                  unit_values=self.unit_values)
        return tmp_sf

    @property
    def mask(self):
        return self.__mask

    @mask.setter
    def mask(self, new_mask):
        # check 'new_mask' coherence
        if isinstance(new_mask, bool):
            fill_value = new_mask
            new_mask = np.empty(self.shape, dtype=bool)
            new_mask.fill(fill_value)
        elif isinstance(new_mask, ARRAYTYPES):
            if not isinstance(new_mask.flat[0], np.bool_):
                raise TypeError()
            new_mask = np.array(new_mask, dtype=bool)
        else:
            raise TypeError("'mask' should be an array or a boolean,"
                            " not a {}".format(type(new_mask)))
        if self.shape != new_mask.shape:
            raise ValueError()
        # check if the new mask don'r reveal masked values
        if np.any(np.logical_not(new_mask[self.mask])):
            raise Warning("This mask reveal masked values, maybe you should"
                          "use the 'fill' function instead")
        # nanify masked values
        self.comp_x[new_mask] = np.NaN
        self.comp_y[new_mask] = np.NaN
        # store mask
        self.__mask = new_mask

    @mask.deleter
    def mask(self):
        raise Exception("Nope, can't do that")

    @property
    def mask_as_sf(self):
        tmp_sf = sf.ScalarField()
        tmp_sf.import_from_arrays(self.axe_x, self.axe_y, self.mask,
                                  mask=False, unit_x=self.unit_x,
                                  unit_y=self.unit_y,
                                  unit_values=self.unit_values)
        return tmp_sf

    @property
    def unit_values(self):
        return self.__unit_values

    @unit_values.setter
    def unit_values(self, new_unit_values):
        if isinstance(new_unit_values, unum.Unum):
            if new_unit_values.asNumber() == 1:
                self.__unit_values = new_unit_values
            else:
                raise ValueError()
        elif isinstance(new_unit_values, STRINGTYPES):
            self.__unit_values = make_unit(new_unit_values)
        else:
            raise TypeError()

    @unit_values.deleter
    def unit_values(self):
        raise Exception("Nope, can't do that")

    @property
    def min(self):
        return np.min(self.magnitude[~self.mask])

    @property
    def max(self):
        return np.max(self.magnitude[~self.mask])

    @property
    def magnitude(self):
        """
        Return a scalar field with the velocity field magnitude.
        """
        comp_x, comp_y = self.comp_x, self.comp_y
        mask = self.mask
        values = (comp_x**2 + comp_y**2)**(.5)
        values[mask] = np.NaN
        return values

    @property
    def magnitude_as_sf(self):
        """
        Return a scalarfield with the velocity field magnitude.
        """
        tmp_sf = sf.ScalarField()
        tmp_sf.import_from_arrays(self.axe_x, self.axe_y, self.magnitude,
                                  mask=self.mask, unit_x=self.unit_x,
                                  unit_y=self.unit_y,
                                  unit_values=self.unit_values)
        return tmp_sf

    @property
    def theta(self):
        """
        Return a scalar field with the vector angle (in reference of the unit_y
        vector [1, 0]).

        Parameters
        ----------
        low_velocity_filter : number
            If not zero, points where V < Vmax*low_velocity_filter are masked.

        Returns
        -------
        theta_sf : sf.ScalarField object
            Contening theta field.
        """
        # get data
        comp_x, comp_y = self.comp_x, self.comp_y
        not_mask = np.logical_not(self.mask)
        theta = np.zeros(self.shape)
        # getting angle
        norm = self.magnitude
        not_mask = np.logical_and(not_mask, norm != 0)
        theta[not_mask] = comp_x[not_mask]/norm[not_mask]
        theta[not_mask] = np.arccos(theta[not_mask])
        tmp_comp_y = comp_y.copy()
        tmp_comp_y[~not_mask] = 0
        sup_not_mask = tmp_comp_y < 0
        theta[sup_not_mask] = 2*np.pi - theta[sup_not_mask]
        return theta

    @property
    def theta_as_sf(self):
        """
        Return a scalarfield with the velocity field angles.
        """
        tmp_sf = sf.ScalarField()
        tmp_sf.import_from_arrays(self.axe_x, self.axe_y, self.theta,
                                  mask=False, unit_x=self.unit_x,
                                  unit_y=self.unit_y,
                                  unit_values=self.unit_values)
        return tmp_sf

    def import_from_arrays(self, axe_x, axe_y, comp_x, comp_y, mask=False,
                           unit_x="", unit_y="", unit_values="", dtype=float):
        """
        Set the vector field from a set of arrays.

        Parameters
        ----------
        axe_x : array
            Discretized axis value along x
        axe_y : array
            Discretized axis value along y
        comp_x : array or masked array
            Values of the x component at the discritized points
        comp_y : array or masked array
            Values of the y component at the discritized points
        mask : array of boolean, optional
            Mask on comp_x and comp_y
        unit_x : string, optionnal
            Unit for the values of axe_x
        unit_y : string, optionnal
            Unit for the values of axe_y
        unit_values : string, optionnal
            Unit for the field components.
        dtype: Numerical type
            Numerical type to store the data to.
            Should be a type supported by numpy arrays.
            Default to 'float'.
        """
        # overwrite previous
        self.__clean()
        # Use numpy arrays
        axe_x = np.array(axe_x, dtype=float)
        axe_y = np.array(axe_y, dtype=float)
        comp_x = np.array(comp_x, dtype=dtype)
        comp_y = np.array(comp_y, dtype=dtype)
        if mask is not None and not isinstance(mask, bool):
            mask = np.array(mask, dtype=bool)
        # Be sure nan values are masked
        mask = np.logical_or(mask, np.isnan(comp_x))
        mask = np.logical_or(mask, np.isnan(comp_y))
        # Be sure axes are one-dimensional
        if axe_x.ndim >= 2:
            if np.all(axe_x[0, 0] == axe_x[:, 0]):
                axe_x = axe_x[0]
            else:
                axe_x = axe_x[:, 0]
        if axe_y.ndim >= 2:
            if np.all(axe_y[0, 0] == axe_y[:, 0]):
                axe_y = axe_y[0]
            else:
                axe_y = axe_y[:, 0]
        # Check if axis are evenly spaced
        delta_x = axe_x[1:] - axe_x[:-1]
        delta_y = axe_y[1:] - axe_y[:-1]
        epsx_abs = delta_x*1e-6
        epsy_abs = delta_y*1e-6
        if np.any(delta_y - delta_y[0] > epsy_abs) or \
           np.any(delta_x - delta_x[0] > epsx_abs):
            warnings.warn("Axis are not evenly spaced.\n"
                          "Consider using 'make_evenly_spaced' method"
                          " to avoid bad surprises.")
        # Be sure comp_x and comp_y are of the good shape
        if len(axe_x) == comp_x.shape[1] and \
           len(axe_y) == comp_x.shape[0] and \
           len(axe_x) != len(axe_y):
            comp_x = comp_x.transpose()
            comp_y = comp_y.transpose()
            try:
                mask = mask.transpose()
            except:
                pass
        # Be sure axes are crescent
        if axe_x[0] > axe_x[-1]:
            axe_x = axe_x[::-1]
            comp_x = comp_x[::-1]
            comp_y = comp_y[::-1]
        if axe_y[0] > axe_y[-1]:
            axe_y = axe_y[::-1]
            comp_x = comp_x[:, ::-1]
            comp_y = comp_y[:, ::-1]
        # Store data
        self.axe_x = axe_x
        self.axe_y = axe_y
        self.comp_x = comp_x
        self.comp_y = comp_y
        self.mask = mask
        self.unit_x = unit_x
        self.unit_y = unit_y
        self.unit_values = unit_values
        self._values_dtype = dtype

    def import_from_sfs(self, vx, vy):
        """
        Set the vector field from two scalarfields.
        """
        # overwrite previous
        self.__clean()
        # Be sure axes matches
        if not np.all(vx.axe_x == vy.axe_x) or not np.all(vx.axe_y == vy.axe_y):
            raise ValueError('Axes do not match')
        # Be sure units match
        if vx.unit_x != vy.unit_x or vx.unit_y != vy.unit_y:
            raise ValueError('Axes units do not match')
        # Be sure units match
        if vx.unit_values != vy.unit_values:
            raise ValueError('Data units do not match')
        # Combine masks
        mask = np.logical_and(vx.mask, vy.mask)
        # Store data
        self.axe_x = vx.axe_x
        self.axe_y = vx.axe_y
        self.comp_x = vx.values
        self.comp_y = vy.values
        self.mask = mask
        self.unit_x = vx.unit_x
        self.unit_y = vx.unit_y
        self.unit_values = vx.unit_values
        self._values_dtype = vx._values_dtype

    def check_consistency(self):
        """
        Raise errors if the vectorfield show some
        incoherences.
        """
        super(VectorField, self).check_consistency()
        # types
        if not isinstance(self.comp_x, np.ndarray):
            raise Exception()
        if not isinstance(self.comp_y, np.ndarray):
            raise Exception()
        if not isinstance(self.mask, np.ndarray):
            raise Exception()
        if not isinstance(self.unit_values, unum.Unum):
            raise Exception()
        # shapes
        shape = (len(self.axe_x), len(self.axe_y))
        if not np.all(self.comp_x.shape == shape):
            raise Exception()
        if not np.all(self.comp_y.shape == shape):
            raise Exception()
        if not np.all(self.mask.shape == shape):
            raise Exception()
        # null values
        if np.any(np.isnan(self.comp_x[~self.mask])):
            raise Exception()
        if np.any(np.isnan(self.comp_y[~self.mask])):
            raise Exception()


    def get_props(self):
        """
        Print the VectorField main properties
        """
        text = "Shape: {}".format(self.shape)
        unit_x = self.unit_x.strUnit()
        text += "Axe x: [{}..{}]{}".format(self.axe_x[0], self.axe_x[-1],
                                           unit_x)
        unit_y = self.unit_y.strUnit()
        text += "Axe y: [{}..{}]{}".format(self.axe_y[0], self.axe_y[-1],
                                           unit_y)
        unit_values = self.unit_values.strUnit()
        xmin = np.min(self.comp_x[~self.mask])
        xmax = np.max(self.comp_x[~self.mask])
        ymin = np.min(self.comp_y[~self.mask])
        ymax = np.max(self.comp_y[~self.mask])
        text += "Comp x: [{}..{}]{}".format(xmin, xmax, unit_values)
        text += "Comp y: [{}..{}]{}".format(ymin, ymax, unit_values)
        nmb_mask = np.sum(self.mask)
        nmb_tot = self.shape[0]*self.shape[1]
        text += "Masked values: {}/{}".format(nmb_mask, nmb_tot)
        return text


    def get_value(self, x, y, ind=False, unit=False):
        """
        Return the vector field components on the point (x, y).
        If ind is true, x and y are indices,
        else, x and y are value on axes (interpolated if necessary).
        """
        return np.array([self.comp_x_as_sf.get_value(x, y,
                                                     ind=ind, unit=unit),
                         self.comp_y_as_sf.get_value(x, y,
                                                     ind=ind, unit=unit)])

    def get_profile(self, component, direction, position, ind=False,
                    interp='linear'):
        """
        Return a profile of the vector field component, at the given position
        (or at least at the nearest possible position).
        If position is an interval, the fonction return an average profile
        in this interval.

        Parameters
        ----------
        component : string in ['vx', 'vy']
            component to treat.
        direction : string in ['x', 'y']
            Direction along which we choose a position.
        position : float or interval of float
            Position or interval in which we want a profile.
        ind : boolean, optional
            If 'True', position is taken as an indice
            Else (default), position is in the field units.
        interp : string in ['nearest', 'linear']
            if 'nearest', get the profile at the nearest position on the grid,
            if 'linear', use linear interpolation to get the profile at the
            exact position

        Returns
        -------
        profile : Profile object
            Asked profile.
        cutposition : array or number
            Final position or interval in which the profile has been taken.
        """
        if component == 'vx':
            return self.comp_x_as_sf.get_profile(direction, position, ind,
                                                 interp=interp)
        elif component == 'vy':
            return self.comp_y_as_sf.get_profile(direction, position, ind,
                                                 interp=interp)
        else:
            raise TypeError("'component' must be 'vx' or 'vy'")

    def copy(self):
        """
        Return a copy of the vectorfield.
        """
        return copy.deepcopy(self)

    def get_norm(self, norm=2, normalized='perpoint'):
        """
        Return the field norm
        """
        values = np.concatenate((self.comp_x[~self.mask],
                                 self.comp_y[~self.mask]))
        res = (np.sum(np.abs(values)**norm))**(1./norm)
        if normalized == "perpoint":
            res /= np.sum(~self.mask)
        return res

    def scale(self, scalex=None, scaley=None, scalev=None, inplace=False):
        """
        Scale the VectorField.

        Parameters
        ----------
        scalex, scaley, scalev : numbers or Unum objects
            Scale for the axis and the values.
        inplace : boolean
            .
        """
        if inplace:
            tmp_f = self
        else:
            tmp_f = self.copy()
        # xy
        revx, revy = field.Field.scale(tmp_f, scalex=scalex, scaley=scaley,
                                       inplace=True, output_reverse=True)
        # v
        if scalev is None:
            pass
        elif isinstance(scalev, NUMBERTYPES):
            tmp_f.comp_x *= scalev
            tmp_f.comp_y *= scalev
        elif isinstance(scalev, unum.Unum):
            new_unit = tmp_f.unit_values*scalev
            fact = new_unit.asNumber()
            new_unit /= fact
            tmp_f.unit_values = new_unit
            tmp_f.comp_x *= fact
            tmp_f.comp_y *= fact
        else:
            raise TypeError()
        if revx and revy:
            tmp_f.comp_x = -tmp_f.comp_x[::-1, ::-1]
            tmp_f.comp_y = -tmp_f.comp_y[::-1, ::-1]
        elif revx:
            tmp_f.comp_x = -tmp_f.comp_x[::-1, :]
            tmp_f.comp_y = tmp_f.comp_y[::-1, :]
        elif revy:
            tmp_f.comp_x = tmp_f.comp_x[:, ::-1]
            tmp_f.comp_y = -tmp_f.comp_y[:, ::-1]
        # returning
        if not inplace:
            return tmp_f

    def rotate(self, angle, inplace=False):
        """
        Rotate the vector field.

        Parameters
        ----------
        angle : integer
            Angle in degrees (positive for trigonometric direction).
            In order to preserve the orthogonal grid, only multiples of
            90° are accepted (can be negative multiples).
        inplace : boolean, optional
            If 'True', vector field is rotated in place, else, the function
            return a rotated field.

        Returns
        -------
        rotated_field : VectorField object, optional
            Rotated vector field.
        """
        # check params
        if not isinstance(angle, NUMBERTYPES):
            raise TypeError()
        if angle % 90 != 0:
            raise ValueError()
        if not isinstance(inplace, bool):
            raise TypeError()
        # get data
        if inplace:
            tmp_field = self
        else:
            tmp_field = self.copy()
        # normalize angle
        angle = angle % 360
        # rotate the parent
        field.Field.rotate(tmp_field, angle, inplace=True)
        # rotate
        nmb_rot90 = int(angle/90)
        comp_x = np.rot90(tmp_field.comp_x, nmb_rot90)
        comp_y = np.rot90(tmp_field.comp_y, nmb_rot90)
        mask = np.rot90(tmp_field.mask, nmb_rot90)
        comp_x2 = np.cos(angle/180.*np.pi)*comp_x - \
                  np.sin(angle/180.*np.pi)*comp_y
        comp_y2 = np.cos(angle/180.*np.pi)*comp_y + \
                  np.sin(angle/180.*np.pi)*comp_x
        tmp_field.__comp_x, tmp_field.__comp_y = comp_x, comp_y
        tmp_field.__comp_x = comp_x2
        tmp_field.__comp_y = comp_y2
        tmp_field.__mask = mask
        # returning
        if not inplace:
            return tmp_field

    def change_dtype(self, new_type):
        """
        Change the values dtype.
        """
        if new_type != self._values_dtype:
            self.comp_x = np.asarray(self.comp_x, dtype=new_type)
            self.comp_y = np.asarray(self.comp_y, dtype=new_type)
            self._values_dtype = new_type

    def change_unit(self, axe, new_unit):
        """
        Change the unit of an axe.

        Parameters
        ----------
        axe : string
            'y' for changing the profile y axis unit
            'x' for changing the profile x axis unit
            'values' or changing values unit
        new_unit : Unum.unit object or string
            The new unit.
        """
        if isinstance(new_unit, STRINGTYPES):
            new_unit = make_unit(new_unit)
        if not isinstance(new_unit, unum.Unum):
            raise TypeError()
        if not isinstance(axe, STRINGTYPES):
            raise TypeError()
        if axe == 'x':
            field.Field.change_unit(self, axe, new_unit)
        elif axe == 'y':
            field.Field.change_unit(self, axe, new_unit)
        elif axe == 'values':
            old_unit = self.unit_values
            new_unit = old_unit.asUnit(new_unit)
            fact = new_unit.asNumber()
            self.comp_x *= fact
            self.comp_y *= fact
            self.unit_values = new_unit/fact
        else:
            raise ValueError()

    def smooth(self, tos='uniform', size=None, inplace=False, **kw):
        """
        Smooth the vectorfield in place.
        Warning : fill up the field (should be used carefully with masked field
        borders)

        Parameters
        ----------
        tos : string, optional
            Type of smoothing, can be 'uniform' (default) or 'gaussian'
            (See ndimage module documentation for more details)
        size : number, optional
            Size of the smoothing (is radius for 'uniform' and
            sigma for 'gaussian').
            Default is 3 for 'uniform' and 1 for 'gaussian'.
        inplace : boolean, optional
            .
        kw : dic
            Additional parameters for ndimage methods
            (See ndimage documentation)
        """
        if not isinstance(tos, STRINGTYPES):
            raise TypeError("'tos' must be a string")
        if size is None and tos == 'uniform':
            size = 3
        elif size is None and tos == 'gaussian':
            size = 1
        # getting field
        if inplace:
            tmp_vf = self
        else:
            tmp_vf = self.copy()
        # filling up the field before smoothing
        tmp_vf.fill(inplace=True)
        # smoothing
        if tos == "uniform":
            tmp_vf.comp_x = ndimage.uniform_filter(tmp_vf.comp_x, size, **kw)
            tmp_vf.comp_y = ndimage.uniform_filter(tmp_vf.comp_y, size, **kw)
        elif tos == "gaussian":
            tmp_vf.comp_x = ndimage.gaussian_filter(tmp_vf.comp_x, size, **kw)
            tmp_vf.comp_y = ndimage.gaussian_filter(tmp_vf.comp_y, size, **kw)
        else:
            raise ValueError("'tos' must be 'uniform' or 'gaussian'")
        # storing
        if not inplace:
            return tmp_vf

    def make_evenly_spaced(self, interp="linear", res=1):
        """
        Use interpolation to make the field evenly spaced

        Parameters
        ----------
        interp : {‘linear’, ‘cubic’, ‘quintic’}, optional
            The kind of spline interpolation to use. Default is ‘linear’.
        res : number
            Resolution of the resulting field.
            A value of 1 meaning a spatial resolution equal to the smallest
            space along the two axis for the initial field.
        """
        # get data
        vx = self.comp_x_as_sf
        vy = self.comp_y_as_sf
        #
        vx.make_evenly_spaced(interp=interp, res=res)
        vy.make_evenly_spaced(interp=interp, res=res)
        # store
        self.import_from_arrays(vx.axe_x, vx.axe_y, vx.values,
                                vy.values, mask=False,
                                unit_x=self.unit_x,
                                unit_y=self.unit_y,
                                unit_values=self.unit_values)

    def fill(self, kind='linear', value=[0., 0.], inplace=False,
             reduce_tri=True, crop=False):
        """
        Fill the masked part of the array.

        Parameters
        ----------
        kind : string, optional
            Type of algorithm used to fill.
            'value' : fill with the given value
            'nearest' : fill with the nearest value
            'linear' (default): fill using linear interpolation
            (Delaunay triangulation)
            'cubic' : fill using cubic interpolation (Delaunay triangulation)
        value : 2x1 array of numbers
            Values used to fill (for kind='value').
        inplace : boolean, optional
            If 'True', fill the sf.ScalarField in place.
            If 'False' (default), return a filled version of the field.
        reduce_tri : boolean, optional
            If 'True', treatment is used to reduce the triangulation effort
            (faster when a lot of masked values)
            If 'False', no treatment (faster when few masked values)
        crop : boolean, optional
            If 'True', TVF borders are croped before filling.
        """
        # check parameters coherence
        if isinstance(value, NUMBERTYPES):
            value = [value, value]
        if not isinstance(value, ARRAYTYPES):
            raise TypeError()
        value = np.array(value)
        if not value.shape == (2,):
            raise ValueError()
        if crop:
            self.crop_masked_border(hard=False, inplace=True)
        # filling components
        comp_x = self.comp_x_as_sf
        comp_y = self.comp_y_as_sf
        new_comp_x = comp_x.fill(kind=kind, value=value[0], inplace=False,
                                 reduce_tri=reduce_tri)
        new_comp_y = comp_y.fill(kind=kind, value=value[1], inplace=False,
                                 reduce_tri=reduce_tri)
        # returning
        if inplace:
            self.comp_x = new_comp_x.values
            self.comp_y = new_comp_y.values
            mask = np.empty(self.shape, dtype=bool)
            mask.fill(False)
            self.__mask = mask
        else:
            vf = VectorField()
            vf.import_from_arrays(self.axe_x, self.axe_y, new_comp_x.values,
                                  new_comp_y.values, mask=False,
                                  unit_x=self.unit_x, unit_y=self.unit_y,
                                  unit_values=self.unit_values)
            return vf

    def crop(self, intervx=None, intervy=None, ind=False,
             inplace=False):
        """
        Crop the area in respect with given intervals.

        Parameters
        ----------
        intervx : array, optional
            interval wanted along x
        intervy : array, optional
            interval wanted along y
        ind : boolean, optional
            If 'True', intervals are understood as indices along axis.
            If 'False' (default), intervals are understood in axis units.
        inplace : boolean, optional
            If 'True', the field is croped in place.
        """
        if inplace:
            indmin_x, indmax_x, indmin_y, indmax_y = \
                field.Field.crop(self, intervx, intervy, full_output=True,
                                 ind=ind, inplace=True)
            self.__comp_x = self.comp_x[indmin_x:indmax_x + 1,
                                        indmin_y:indmax_y + 1]
            self.__comp_y = self.comp_y[indmin_x:indmax_x + 1,
                                        indmin_y:indmax_y + 1]
            self.__mask = self.mask[indmin_x:indmax_x + 1,
                                    indmin_y:indmax_y + 1]
        else:
            indmin_x, indmax_x, indmin_y, indmax_y, cropfield = \
                field.Field.crop(self, intervx=intervx, intervy=intervy,
                                 full_output=True, ind=ind)
            cropfield.__comp_x = self.comp_x[indmin_x:indmax_x + 1,
                                             indmin_y:indmax_y + 1]
            cropfield.__comp_y = self.comp_y[indmin_x:indmax_x + 1,
                                             indmin_y:indmax_y + 1]
            cropfield.__mask = self.mask[indmin_x:indmax_x + 1,
                                         indmin_y:indmax_y + 1]
            return cropfield

    def crop_masked_border(self, hard=False, inplace=False):
        """
        Crop the masked border of the field in place.

        Parameters
        ----------
        hard : boolean, optional
            If 'True', partially masked border are croped as well.
        """
        #
        if inplace:
            tmp_sf = self
        else:
            tmp_sf = self.copy()
        # checking masked values presence
        mask = tmp_sf.mask
        if not np.any(mask):
            return None
        # hard cropping
        if hard:
            # remove trivial borders
            tmp_sf.crop_masked_border(hard=False, inplace=True)
            # until there is no more masked values
            while np.any(tmp_sf.mask):
                # getting number of masked value on each border
                bd1 = np.sum(tmp_sf.mask[0, :])
                bd2 = np.sum(tmp_sf.mask[-1, :])
                bd3 = np.sum(tmp_sf.mask[:, 0])
                bd4 = np.sum(tmp_sf.mask[:, -1])
                # getting more masked border
                more_masked = np.argmax([bd1, bd2, bd3, bd4])
                # deleting more masked border
                if more_masked == 0:
                    len_x = len(tmp_sf.axe_x)
                    tmp_sf.crop(intervx=[1, len_x], ind=True, inplace=True)
                elif more_masked == 1:
                    len_x = len(tmp_sf.axe_x)
                    tmp_sf.crop(intervx=[0, len_x - 2], ind=True, inplace=True)
                elif more_masked == 2:
                    len_y = len(self.axe_y)
                    tmp_sf.crop(intervy=[1, len_y], ind=True,
                                inplace=True)
                elif more_masked == 3:
                    len_y = len(tmp_sf.axe_y)
                    tmp_sf.crop(intervy=[0, len_y - 2], ind=True, inplace=True)
        # soft cropping
        else:
            axe_x_m = np.logical_not(np.all(mask, axis=1))
            axe_y_m = np.logical_not(np.all(mask, axis=0))
            axe_x_min = np.where(axe_x_m)[0][0]
            axe_x_max = np.where(axe_x_m)[0][-1]
            axe_y_min = np.where(axe_y_m)[0][0]
            axe_y_max = np.where(axe_y_m)[0][-1]
            tmp_sf.crop([axe_x_min, axe_x_max], [axe_y_min, axe_y_max],
                        ind=True, inplace=True)
        # returning
        if not inplace:
            return tmp_sf

    def extend(self, nmb_left=0, nmb_right=0, nmb_up=0, nmb_down=0,
               inplace=False):
        """
        Add columns or lines of masked values at the vectorfield.

        Parameters
        ----------
        nmb_**** : integers
            Number of lines/columns to add in each direction.
        inplace : bool
            If 'False', return a new extended field, if 'True', modify the
            field inplace.
        Returns
        -------
        Extended_field : Field object, optional
            Extended field.
        """
        if inplace:
            field.Field.extend(self, nmb_left=nmb_left, nmb_right=nmb_right,
                               nmb_up=nmb_up, nmb_down=nmb_down, inplace=True)
            new_shape = self.shape
        else:
            new_field = field.Field.extend(self, nmb_left=nmb_left,
                                           nmb_right=nmb_right, nmb_up=nmb_up,
                                           nmb_down=nmb_down, inplace=False)
            new_shape = new_field.shape
        new_Vx = np.zeros(new_shape, dtype=float)
        new_Vy = np.zeros(new_shape, dtype=float)
        if nmb_right == 0:
            slice_x = slice(nmb_left, new_Vx.shape[0] + 2)
        else:
            slice_x = slice(nmb_left, -nmb_right)
        if nmb_up == 0:
            slice_y = slice(nmb_down, new_Vx.shape[1] + 2)
        else:
            slice_y = slice(nmb_down, -nmb_up)
        new_Vx[slice_x, slice_y] = self.comp_x
        new_Vy[slice_x, slice_y] = self.comp_y
        new_mask = np.ones(new_shape, dtype=bool)
        new_mask[slice_x, slice_y] = self.mask
        if inplace:
            self.comp_x = new_Vx
            self.comp_y = new_Vy
            self.__mask = new_mask
        else:
            new_field.comp_x = new_Vx
            new_field.comp_y = new_Vy
            new_field.__mask = new_mask
            return new_field

    def mirroring(self, direction, position, inds_to_mirror='all', mir_coef=1.,
                  inplace=False, interp=None, value=[0, 0]):
        """
        Return a field with additional mirrored values.

        Parameters
        ----------
        direction : string in ['x', 'y']
            Axe on which place the symetry plane.
        position : number
            Position of the symetry plane along the given axe
        inds_to_mirror : integer
            Number of vector rows to symetrize (default is all)
        mir_coef : number or 2x1 array, optional
            Optional coefficient(s) applied only to the mirrored values.
            If ana array first value is for 'comp_x' and second one to 'comp_y'
        inplace : boolean, optional
            .
        interp : string, optional
            If specified, method used to fill the gap near the
            symetry plane by interpoaltion.
            'value' : fill with the given value,
            'nearest' : fill with the nearest value,
            'linear' (default): fill using linear interpolation
            (Delaunay triangulation),
            'cubic' : fill using cubic interpolation (Delaunay triangulation)
        value : array, optional
            Value at the symetry plane, in case of interpolation
        """
        if direction not in ['x', 'y']:
            raise ValueError()
        # getting components
        vx = self.comp_x_as_sf
        vy = self.comp_y_as_sf
        xy_scale = self.xy_scale
        # treating sign changments
        if isinstance(mir_coef, NUMBERTYPES):
            if direction == 'x':
                coefx = -1
                coefy = 1
            else:
                coefx = 1
                coefy = -1
            coefx *= mir_coef
            coefy *= mir_coef
        elif isinstance(mir_coef, ARRAYTYPES):
            coefx = mir_coef[0]
            coefy = mir_coef[1]
        else:
            raise ValueError()
        # mirroring on components
        vx.mirroring(direction, position, inds_to_mirror=inds_to_mirror,
                     interp=interp, value=value[0], inplace=True,
                     mir_coef=coefx)
        vy.mirroring(direction, position, inds_to_mirror=inds_to_mirror,
                     interp=interp, value=value[1], inplace=True,
                     mir_coef=coefy)
        # storing
        if inplace:
            tmp_vf = self
        else:
            tmp_vf = VectorField()
        mask = np.logical_or(vx.mask, vy.mask)
        tmp_vf.import_from_arrays(vx.axe_x, vx.axe_y, vx.values, vy.values,
                                  mask=mask, unit_x=vx.unit_x,
                                  unit_y=vy.unit_y,
                                  unit_values=vx.unit_values)
        tmp_vf.xy_scale = xy_scale
        # returning
        if not inplace:
            return tmp_vf

    def reduce_spatial_resolution(self, fact, inplace=False):
        """
        Reduce the spatial resolution of the field by a factor 'fact'

        Parameters
        ----------
        fact : int
            Reducing factor.
        inplace : boolean, optional
            .
        """
        # reducing
        Vx = self.comp_x_as_sf
        Vy = self.comp_y_as_sf
        Vx.reduce_spatial_resolution(fact, inplace=True)
        Vy.reduce_spatial_resolution(fact, inplace=True)
        # returning
        if inplace:
            self.__init__()
            self.import_from_arrays(Vx.axe_x, Vx.axe_y, Vx.values,
                                    Vy.values,
                                    mask=Vx.mask, unit_x=self.unit_x,
                                    unit_y=self.unit_y,
                                    unit_values=self.unit_values)
        else:
            vf = VectorField()
            vf.import_from_arrays(Vx.axe_x, Vx.axe_y, Vx.values, Vy.values,
                                  mask=Vx.mask, unit_x=self.unit_x,
                                  unit_y=self.unit_y,
                                  unit_values=self.unit_values)
            return vf

    def __clean(self):
        self.__init__()

    def _display(self, component=None, kind=None, axis='image', **plotargs):
        # adapt xy scale if streamplot is needed
        if kind == "stream" and self.xy_scale.asNumber() != 1.:
            magn = self.magnitude
            new_comp_x = self.comp_x*self.xy_scale.asNumber()
            new_magn = (new_comp_x**2 + self.comp_y**2)**.5
            values = [new_comp_x/new_magn*magn, self.comp_y/new_magn*magn]
        elif component in [None, 'V']:
            values = [self.comp_x, self.comp_y]
        elif component in ['comp_x', 'x']:
            values = self.comp_x
        elif component in ['comp_y', 'y']:
            values = self.comp_y
        else:
            values = self.__getattribute__(component)
        dp = pplt.Displayer(x=self.axe_x, y=self.axe_y, values=values,
                            kind=kind, **plotargs)
        plot = dp.draw(cb=False)
        pplt.DataCursorTextDisplayer(dp)
        unit_x, unit_y = self.unit_x, self.unit_y
        if unit_x.strUnit() == "[]":
            plt.xlabel("x")
        else:
            plt.xlabel("x " + unit_x.strUnit())
        if unit_y.strUnit() == "[]":
            plt.ylabel("y")
        else:
            plt.ylabel("y " + unit_y.strUnit())
        return plot

    def display(self, component=None, kind=None, **plotargs):
        """
        Display something from the vector field.
        If component is not given, a quiver is displayed.
        If component is an integer, the coresponding component of the field is
        displayed.

        Parameters
        ----------
        component : string, optional
            Component to display, can be 'V', 'x', 'y', 'mask'
        kind : string, optinnal
            Scalar plots :
            if 'None': each datas are plotted (imshow),
            if 'contour': contours are ploted  (contour),
            if 'contourf': filled contours are ploted (contourf).
            Vector plots :
            if 'quiver': quiver plot,
            if 'stream': streamlines.

        plotargs : dict
            Arguments passed to the function used to display the vector field.

        Returns
        -------
        fig : figure reference
            Reference to the displayed figure
        """
        displ = self._display(component, kind, **plotargs)
        unit_values = self.unit_values
        # Vx, Vy = self.comp_x, self.comp_y
        if component is None or component == 'V':
            if kind == 'quiver' or kind is None:
                if 'C' not in list(plotargs.keys()):
                    cb = plt.colorbar(displ)
                    if unit_values.strUnit() == "[]":
                        cb.set_label("Magnitude")
                    else:
                        cb.set_label("Magnitude " + unit_values.strUnit())
                # legendarrow = round(np.max([Vx.max(), Vy.max()]))
                # plt.quiverkey(displ, 1.075, 1.075, legendarrow,
                #               "$" + str(legendarrow)
                #               + unit_values.strUnit() + "$",
                #               labelpos='W', fontproperties={'weight': 'bold'})
            elif kind in ['stream', 'track']:
                if 'color' not in list(plotargs.keys()):
                    cb = plt.colorbar(displ.lines)
                    if unit_values.strUnit() == "[]":
                        cb.set_label("Magnitude")
                    else:
                        cb.set_label("Magnitude " + unit_values.strUnit())
        elif component in ['x', 'comp_x']:
            cb = plt.colorbar(displ)
            if unit_values.strUnit() == "[]":
                cb.set_label("Vx")
            else:
                cb.set_label("Vx " + unit_values.strUnit())
        elif component in ['y', 'comp_y']:
            cb = plt.colorbar(displ)
            if unit_values.strUnit() == "[]":
                cb.set_label("Vy")
            else:
                cb.set_label("Vy " + unit_values.strUnit())
        elif component == 'mask':
            cb = plt.colorbar(displ)
            cb.set_label("Mask")
        elif component == 'magnitude':
            cb = plt.colorbar(displ)
            if unit_values.strUnit() == "[]":
                cb.set_label("Magnitude")
            else:
                cb.set_label("Magnitude " + unit_values.strUnit())
        else:
            raise ValueError("Unknown 'component' value")
        plt.title("")
        return displ



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from .field_treatment import get_streamlines,\
    get_shear_stress, reconstruct_from_gradients, \
    get_swirling_vector, get_gradients, get_grad_field, \
    get_jacobian_eigenproperties, get_Kenwright_field, \
    get_tracklines, get_track_field, extrapolate_until_wall, \
    get_streamlines_fast, get_divergence, get_fieldlines



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os
import sys
try:
    dirname = os.path.dirname(os.path.realpath(__file__))
    sys.path.append(dirname)
    os.chdir(dirname)
except:
    pass
import matplotlib as mpl
mpl.use('Agg')

import numpy as np
import pytest
from helper import sane_parameters, parametric_test

import matplotlib.pyplot as plt
from IMTreatment import VectorField, file_operation as imtio, make_unit


class TestVectorField(object):
    """ Done """

    def setup(self):
        sane_parameters()
        self.VF1 = imtio.import_from_file("VF1.cimt")
        self.VF1_nomask = imtio.import_from_file("VF1_nomask.cimt")
        self.VF2 = imtio.import_from_file("VF2.cimt")
        self.VF_notevenlyspaced = imtio.import_from_file("VF_notevenlyspaced.cimt")

    def test_import_from_arrays(self):
        # creating a VF field using 'import_from_arrays
        axe_x = np.arange(10)
        unit_x = make_unit('m')
        unit_y = make_unit('km')
        unit_values = make_unit('m/s')
        axe_y = np.arange(20)*.01
        vx = np.arange(len(axe_x)*len(axe_y)).reshape(len(axe_y), len(axe_x))
        vx = np.array(vx, dtype=float)
        vy = np.random.rand(len(axe_y), len(axe_x))
        mask = np.random.rand(len(axe_y), len(axe_x)) > 0.75
        vx[mask] = np.nan
        vx = np.pi*vx
        vy[mask] = np.nan
        vy = np.pi*vy
        vf = VectorField()
        vf.import_from_arrays(axe_x, axe_y, vx, vy, mask=mask,
                              unit_x=unit_x, unit_y=unit_y,
                              unit_values=unit_values)
        vx = vx.transpose()
        vy = vy.transpose()
        mask = mask.transpose()
        # tests
        assert np.all(vf.axe_x == axe_x)
        assert np.all(vf.axe_y == axe_y)
        assert np.all(vf.comp_x[~vf.mask] == vx[~mask])
        assert np.all(vf.comp_y[~vf.mask] == vy[~mask])
        assert np.all(vf.mask == mask)
        assert vf.unit_x == unit_x
        assert vf.unit_y == unit_y
        assert vf.unit_values == unit_values

    def test_operations(self):
        # get datas
        axe_x, axe_y = self.VF1.axe_x, self.VF1.axe_y
        vx = self.VF1.comp_x
        vy = self.VF1.comp_y
        mask = self.VF1.mask
        vx2 = self.VF2.comp_x
        vy2 = self.VF2.comp_y
        mask2 = self.VF2.mask
        unit_x, unit_y = self.VF1.unit_x, self.VF1.unit_y
        unit_values = self.VF1.unit_values
        # neg
        vf = -self.VF1.copy()
        assert np.all(vf.axe_x == axe_x)
        assert np.all(vf.axe_y == axe_y)
        assert np.all(vf.comp_x[~vf.mask] == -vx[~mask])
        assert np.all(vf.comp_y[~vf.mask] == -vy[~mask])
        assert np.all(vf.mask == mask)
        assert vf.unit_x == unit_x
        assert vf.unit_y == unit_y
        assert vf.unit_values == unit_values
        # add
        nmb = 5
        unt = 500*make_unit('mm/s')
        vx_f = (nmb + vx + unt.asNumber()/1000. + vx2 +
                unt.asNumber()/1000. + vx + nmb)
        vy_f = (nmb + vy + unt.asNumber()/1000. + vy2 +
                unt.asNumber()/1000. + vy + nmb)
        mask_f = np.logical_or(mask, mask2)
        vf = nmb + self.VF1 + unt + self.VF2 + unt + self.VF1 + nmb
        assert np.all(vf.axe_x == axe_x)
        assert np.all(vf.axe_y == axe_y)
        assert np.all(vf.comp_x[~mask_f] == vx_f[~mask_f])
        assert np.all(vf.comp_y[~mask_f] == vy_f[~mask_f])
        assert np.all(vf.mask == mask_f)
        assert vf.unit_x == unit_x
        assert vf.unit_y == unit_y
        assert vf.unit_values == unit_values
        # sub
        nmb = 5
        unt = 500*make_unit('mm/s')
        vx_f = (nmb - vx - unt.asNumber()/1000. - vx2 -
                unt.asNumber()/1000. - vx - nmb)
        vy_f = (nmb - vy - unt.asNumber()/1000. - vy2 -
                unt.asNumber()/1000. - vy - nmb)
        mask_f = np.logical_or(mask, mask2)
        vf = nmb - self.VF1 - unt - self.VF2 - unt - self.VF1 - nmb
        assert np.all(vf.axe_x == axe_x)
        assert np.all(vf.axe_y == axe_y)
        assert np.all(vf.comp_x[~mask_f] == vx_f[~mask_f])
        assert np.all(vf.comp_y[~mask_f] == vy_f[~mask_f])
        assert np.all(vf.mask == mask_f)
        assert vf.unit_x == unit_x
        assert vf.unit_y == unit_y
        assert vf.unit_values == unit_values
        # mul
        nmb = 5.23
        unt = 500.*make_unit('mm/s')
        unt_n = 500./1000.
        vx_f = (nmb * vx * unt_n + vx2 *
                unt_n * nmb)
        vy_f = (nmb * vy * unt_n + vy2 *
                unt_n * nmb)
        unit_values = make_unit('mm/s')*make_unit('m/s')*1e3
        mask_f = np.logical_or(mask, mask2)
        vf = nmb * self.VF1 * unt + self.VF2 * unt * nmb
        assert np.all(vf.axe_x == axe_x)
        assert np.all(vf.axe_y == axe_y)
        assert np.all(vf.mask == mask_f)
        assert np.all(vf.comp_x[~vf.mask] - vx_f[~mask_f] < 1e-6)
        assert np.all(vf.comp_y[~vf.mask] - vy_f[~mask_f] < 1e-6)
        assert vf.unit_x == unit_x
        assert vf.unit_y == unit_y
        assert vf.unit_values == unit_values
        # div
        nmb = 5.23
        unt = 500.*make_unit('kg*mm/s')
        unt_n = 500./1000.
        vx_f = nmb / (vx / unt_n) + 1./(vx2 / unt_n)
        vy_f = nmb / (vy / unt_n) + 1./(vy2 / unt_n)
        unit_values = 1./(make_unit('m/s')/make_unit('kg*mm/s'))*1e3
        mask_f = np.logical_or(mask, mask2)
        vf = nmb / (self.VF1 / unt) + 1./(self.VF2 / unt)
        assert np.all(vf.axe_x == axe_x)
        assert np.all(vf.axe_y == axe_y)
        assert np.all(vf.mask == mask_f)
        assert np.all(vf.comp_x[~vf.mask] - vx_f[~mask_f] < 1e-6)
        assert np.all(vf.comp_y[~vf.mask] - vy_f[~mask_f] < 1e-6)
        assert vf.unit_x == unit_x
        assert vf.unit_y == unit_y
        assert vf.unit_values == unit_values
        # abs
        unit_values = self.VF1.unit_values
        vf = np.abs(self.VF1)
        assert np.all(vf.axe_x == axe_x)
        assert np.all(vf.axe_y == axe_y)
        assert np.all(vf.comp_x[~vf.mask] == np.abs(vx[~mask]))
        assert np.all(vf.comp_y[~vf.mask] == np.abs(vy[~mask]))
        assert np.all(vf.mask == mask)
        assert vf.unit_x == unit_x
        assert vf.unit_y == unit_y
        assert vf.unit_values == unit_values
        # pow
        unit_values = self.VF1.unit_values**3.544186
        vf = (np.abs(self.VF1) + 1)**3.544186
        assert np.all(vf.axe_x == axe_x)
        assert np.all(vf.axe_y == axe_y)
        assert np.all(vf.comp_x[~vf.mask] -
                      (np.abs(vx[~mask]) + 1)**3.544186 <
                      1e-6)
        assert np.all(vf.comp_y[~vf.mask] -
                      (np.abs(vy[~mask]) + 1)**3.544186 <
                      1e-6)
        assert np.all(vf.mask == mask)
        assert vf.unit_x == unit_x
        assert vf.unit_y == unit_y
        assert vf.unit_values == unit_values

    def test_iter(self):
        res = list(self.VF1.__iter__())
        # imtio.export_to_file(res, "VF1_iter.cimt")
        res2 = imtio.import_from_file("VF1_iter.cimt")
        assert res == res2

    def test_trim_area(self):
        axe_x, axe_y = self.VF1.axe_x, self.VF1.axe_y
        vx = self.VF1.comp_x
        vy = self.VF1.comp_y
        mask = self.VF1.mask
        vf = self.VF1.crop([axe_x[3], axe_x[-4]], [axe_y[2], axe_y[-7]])
        assert np.all(vf.axe_x == axe_x[3:-3])
        assert np.all(vf.axe_y == axe_y[2:-6])
        assert np.all(vf.comp_x[~vf.mask] ==
                      vx[3:-3, 2:-6][~mask[3:-3, 2:-6]])
        assert np.all(vf.comp_y[~vf.mask] ==
                      vy[3:-3, 2:-6][~mask[3:-3, 2:-6]])

    def test_min_max(self):
        mini = self.VF1.min
        maxi = self.VF1.max
        assert mini == 0.54631058701128665
        assert maxi == 97.570083275983819

    def test_get_value(self):
        res = self.VF1.get_value(7.5, 20.12)
        # imtio.export_to_file(res, "VF1_get_value.cimt")
        res2 = imtio.import_from_file("VF1_get_value.cimt")
        assert np.all(res == res2)
        res_b = self.VF1.get_value(5, 10, ind=True)
        # imtio.export_to_file(res_b, "VF1_get_value_b.cimt")
        res_b2 = imtio.import_from_file("VF1_get_value_b.cimt")
        assert np.all(res_b == res_b2)

    def test_magnitude(self):
        res = self.VF1.magnitude
        # imtio.export_to_file(res, "VF1_magnitude.cimt")
        res2 = imtio.import_from_file("VF1_magnitude.cimt")
        assert np.all(res[~np.isnan(res)] == res2[~np.isnan(res2)])

    def test_magnitude_as_sf(self):
        res = self.VF1.magnitude_as_sf
        # imtio.export_to_file(res, "VF1_magnitude_as_sf.cimt")
        res2 = imtio.import_from_file("VF1_magnitude_as_sf.cimt")
        assert res == res2

    def test_theta(self):
        res = self.VF1.theta
        # imtio.export_to_file(res, "VF1_theta.cimt")
        res2 = imtio.import_from_file("VF1_theta.cimt")
        assert np.all(res == res2)

    def test_theta_as_sf(self):
        res = self.VF1.theta_as_sf
        # imtio.export_to_file(res, "VF1_theta_as_sf.cimt")
        res2 = imtio.import_from_file("VF1_theta_as_sf.cimt")
        assert res == res2

    def test_get_profile(self):
        res = self.VF1.get_profile('vx', 'x', 7.2)
        # imtio.export_to_file(res, "VF1_get_profile.cimt")
        res2 = imtio.import_from_file("VF1_get_profile.cimt")
        assert res == res2
        res_b = self.VF1.get_profile('vy', 'x', 3, ind=True)
        # imtio.export_to_file(res_b, "VF1_get_profile_b.cimt")
        res_b2 = imtio.import_from_file("VF1_get_profile_b.cimt")
        assert res_b == res_b2
        res_c = self.VF1.get_profile('vy', 'y', 2.34)
        # imtio.export_to_file(res_c, "VF1_get_profile_c.cimt")
        res_c2 = imtio.import_from_file("VF1_get_profile_c.cimt")
        assert res_c == res_c2
        res_d = self.VF1.get_profile('vx', 'y', 19, ind=True)
        # imtio.export_to_file(res_d, "VF1_get_profile_d.cimt")
        res_d2 = imtio.import_from_file("VF1_get_profile_d.cimt")
        assert res_d == res_d2

    def test_copy(self):
        res_a = self.VF1.copy()
        # imtio.export_to_file(res_a, "VF1_copy_a.cimt")
        res_a2 = imtio.import_from_file("VF1_copy_a.cimt")
        assert res_a == res_a2

    def test_get_norm(self):
        res_a = self.VF1.get_norm()
        # imtio.export_to_file(res_a, "VF1_get_norm_a.cimt")
        res_a2 = imtio.import_from_file("VF1_get_norm_a.cimt")
        assert res_a == res_a2

    def test_scale(self):
        res_a = self.VF1.scale(scalex=1.65, scaley=0.64, scalev=9,
                               inplace=False)
        # imtio.export_to_file(res_a, "VF1_scale_a.cimt")
        res_a2 = imtio.import_from_file("VF1_scale_a.cimt")
        assert res_a == res_a2

    def test_rotate(self):
        res_a = self.VF1.copy()
        res_a.rotate(270)
        # imtio.export_to_file(res_a, "VF1_rotate_a.cimt")
        res_a2 = imtio.import_from_file("VF1_rotate_a.cimt")
        assert res_a == res_a2

    def test_change_unit(self):
        res_a = self.VF1.copy()
        res_a.change_unit("x", "km")
        res_a.change_unit('y', 'um')
        res_a.change_unit('values', 'km/us')
        # imtio.export_to_file(res_a, "VF1_change_unit_a.cimt")
        res_a2 = imtio.import_from_file("VF1_change_unit_a.cimt")
        assert res_a == res_a2

    def test_smooth(self):
        res_a = self.VF1.smooth(tos='uniform', size=7, inplace=False)
        # imtio.export_to_file(res_a, "VF1_smooth_a.cimt")
        res_a2 = imtio.import_from_file("VF1_smooth_a.cimt")
        assert res_a == res_a2

    def test_make_evenly_spaced(self):
        res_a = self.VF_notevenlyspaced.make_evenly_spaced()
        # imtio.export_to_file(res_a, "VF1_make_evenly_spaced_a.cimt")
        res_a2 = imtio.import_from_file("VF1_make_evenly_spaced_a.cimt")
        assert res_a == res_a2

    def test_fill(self):
        res_a = self.VF1.fill(inplace=False)
        # imtio.export_to_file(res_a, "VF1_fill_a.cimt")
        res_a2 = imtio.import_from_file("VF1_fill_a.cimt")
        assert res_a == res_a2

    def test_crop(self):
        res = self.VF1.crop(intervx=[7.4, 19.0],
                            intervy=[2.1, 8], inplace=False)
        # imtio.export_to_file(res, "VF1_crop.cimt")
        res2 = imtio.import_from_file("VF1_crop.cimt")
        assert res == res2

    def test_crop_masked_border(self):
        tmpvf = self.VF1.copy()
        tmpvf.mask[0:3, :] = True
        tmpvf.mask[-1:, :] = True
        tmpvf.mask[:, 0:5] = True
        tmpvf.mask[:, -4:] = True
        res = tmpvf.crop_masked_border(inplace=False)
        # imtio.export_to_file(res, "VF1_crop_masked_border.cimt")
        res2 = imtio.import_from_file("VF1_crop_masked_border.cimt")
        assert res == res2
        #
        tmpvf = self.VF1_nomask.copy()
        tmpvf.mask[3:-2, :] = False
        tmpvf.mask[:, 5:-4] = False
        res = self.VF1.crop_masked_border(hard=True, inplace=False)
        # imtio.export_to_file(res, "VF1_crop_masked_border_2.cimt")
        res2 = imtio.import_from_file("VF1_crop_masked_border_2.cimt")
        assert res == res2

    def test_extend(self):
        res_a = self.VF1.extend(nmb_down=4, nmb_left=2, nmb_right=7, nmb_up=12,
                                inplace=False)
        # imtio.export_to_file(res_a, "VF1_extend_a.cimt")
        res_a2 = imtio.import_from_file("VF1_extend_a.cimt")
        assert res_a == res_a2

    def test_mirroring(self):
        res_a = self.VF1.mirroring('x', 4)
        # imtio.export_to_file(res_a, "VF1_mirroring_a.cimt")
        res_a2 = imtio.import_from_file("VF1_mirroring_a.cimt")
        assert res_a == res_a2
        #
        res_b = self.VF1.copy()
        res_b.mirroring('x', 10, mir_coef=-1.2, inplace=True)
        # imtio.export_to_file(res_b, "VF1_mirroring_b.cimt")
        res_b2 = imtio.import_from_file("VF1_mirroring_b.cimt")
        assert res_b == res_b2

    def test_reduce_spatial_resolution(self):
        res = self.VF1.reduce_spatial_resolution(4, inplace=False)
        # imtio.export_to_file(res, "VF1_reduce_spatial_resolution.cimt")
        res2 = imtio.import_from_file("VF1_reduce_spatial_resolution.cimt")
        assert res == res2

    @pytest.mark.mpl_image_compare
    def test_display_VF_a(self):
        fig = plt.figure()
        self.VF1.display()
        return fig

    @pytest.mark.mpl_image_compare
    def test_display_VF_b(self):
        fig = plt.figure()
        self.VF1.display("magnitude")
        return fig

    @pytest.mark.mpl_image_compare
    def test_display_VF_c(self):
        fig = plt.figure()
        self.VF1.display("mask")
        return fig

    @pytest.mark.mpl_image_compare
    def test_display_VF_d(self):
        fig = plt.figure()
        self.VF1.display("x")
        return fig

    @pytest.mark.mpl_image_compare
    def test_display_VF_e(self):
        fig = plt.figure()
        self.VF1.display("y")
        return fig

    @pytest.mark.mpl_image_compare
    def test_display_VF_f(self):
        fig = plt.figure()
        self.VF1.display(kind='stream')
        return fig

    @pytest.mark.mpl_image_compare
    def test_display_VF_g(self):
        fig = plt.figure()
        self.VF1.display(kind='stream', density=2, color='k')
        return fig

# TEMP
pytest.main(["test_vectorfield.py"])
# TEMP - End
